import os
from pathlib import Path

import pytest

from wsireg.reg_transform_seq import RegTransformSeq

HERE = os.path.dirname(__file__)
FIXTURES_DIR = os.path.join(HERE, "fixtures")


@pytest.mark.usefixtures("complex_transform_larger")
def test_RegTransformSeq_from_dict(complex_transform_larger):
    rts = RegTransformSeq(complex_transform_larger)

    assert len(rts.reg_transforms) == 6
    assert len(rts.reg_transforms_itk_order) == 6
    assert rts.transform_seq_idx == [0, 1, 2, 2, 3, 3]


@pytest.mark.usefixtures("complex_transform_larger")
def test_RegTransformSeq_resize_output_up(complex_transform_larger):
    rts = RegTransformSeq(complex_transform_larger)
    os_pre = rts.output_size
    rts.set_output_spacing((1, 1))

    assert rts.output_size == (2048, 2048)
    assert os_pre != rts.output_size


@pytest.mark.usefixtures("complex_transform_larger")
def test_RegTransformSeq_resize_output_down(complex_transform_larger):
    rts = RegTransformSeq(complex_transform_larger)
    os_pre = rts.output_size
    rts.set_output_spacing((4, 4))

    assert rts.output_size == (512, 512)
    assert os_pre != rts.output_size


def test_RegTransformSeq_from_json():
    test_tform = str(Path(FIXTURES_DIR) / "test-tform.json")
    rts = RegTransformSeq(test_tform)

    assert len(rts.reg_transforms) == 10


def test_RegTransformSeq_from_RegTransform():
    test_tform = str(Path(FIXTURES_DIR) / "test-tform.json")
    rts = RegTransformSeq(test_tform)
    rts_1 = RegTransformSeq(
        reg_transforms=[
            rts.reg_transforms[0],
            rts.reg_transforms[1],
            rts.reg_transforms[2],
        ],
        transform_seq_idx=[0, 1, 1],
    )
    assert len(rts_1.reg_transforms) == 3
    assert rts_1.transform_seq_idx == [0, 1, 1]


def test_RegTransformSeq_append():
    test_tform = str(Path(FIXTURES_DIR) / "test-tform.json")
    rts = RegTransformSeq(test_tform)
    rts_1 = RegTransformSeq(
        reg_transforms=[
            rts.reg_transforms[0],
            rts.reg_transforms[1],
            rts.reg_transforms[2],
        ],
        transform_seq_idx=[0, 1, 1],
    )
    rts_2 = RegTransformSeq(
        reg_transforms=[
            rts.reg_transforms[0],
            rts.reg_transforms[1],
            rts.reg_transforms[2],
        ],
        transform_seq_idx=[0, 1, 1],
    )
    rts_1.append(rts_2)

    assert len(rts_1.reg_transforms) == 6
    assert rts_1.transform_seq_idx == [0, 1, 1, 2, 3, 3]



import dask.array as da
import numpy as np
import pytest
import zarr
from tifffile import TiffWriter, imread, imwrite


@pytest.fixture
def im_gry_np():
    return np.random.randint(0, 255, (2048, 2048), dtype=np.uint16)


@pytest.fixture
def mask_np():
    mask_im = np.zeros((2048, 2048), dtype=np.uint8)
    mask_im[256:1792, 256:1792] = 255
    return mask_im


@pytest.fixture
def im_mch_np():
    return np.random.randint(0, 255, (3, 2048, 2048), dtype=np.uint16)


@pytest.fixture
def im_rgb_np():
    return np.random.randint(0, 255, (2048, 2048, 3), dtype=np.uint8)


@pytest.fixture
def dask_im_gry_np():
    return da.from_array(
        np.random.randint(0, 255, (2048, 2048), dtype=np.uint16)
    )


@pytest.fixture
def dask_im_mch_np():
    return da.from_array(
        np.random.randint(0, 255, (3, 2048, 2048), dtype=np.uint16)
    )


@pytest.fixture
def dask_im_rgb_np():
    return da.from_array(
        np.random.randint(0, 255, (2048, 2048, 3), dtype=np.uint8)
    )


@pytest.fixture
def zarr_im_gry_np():
    return zarr.array(np.random.randint(0, 255, (2048, 2048), dtype=np.uint16))


@pytest.fixture
def zarr_im_mch_np():
    return zarr.array(
        np.random.randint(0, 255, (3, 2048, 2048), dtype=np.uint16)
    )


@pytest.fixture
def zarr_im_rgb_np():
    return zarr.array(
        np.random.randint(0, 255, (2048, 2048, 3), dtype=np.uint8)
    )


@pytest.fixture
def disk_im_mch(tmpdir_factory, im_mch_np):
    out_im = tmpdir_factory.mktemp("image").join("image_fp_mch.tiff")
    imwrite(out_im, im_mch_np, tile=(256, 256))
    return out_im


@pytest.fixture
def disk_im_mch_notile(tmpdir_factory, im_mch_np):
    out_im = tmpdir_factory.mktemp("image").join("image_fp_mch_nt.tiff")
    imwrite(out_im, im_mch_np, photometric="MINISBLACK", tile=(256, 256))
    return out_im


@pytest.fixture
def disk_im_rgb(tmpdir_factory, im_rgb_np):
    out_im = tmpdir_factory.mktemp("image").join("image_fp_rgb.tiff")
    imwrite(out_im, im_rgb_np, tile=(256, 256))
    return out_im


@pytest.fixture
def disk_im_gry(tmpdir_factory, im_gry_np):
    out_im = tmpdir_factory.mktemp("image").join("image_fp_gry.tiff")
    imwrite(out_im, im_gry_np, tile=(256, 256))
    return out_im


@pytest.fixture
def disk_im_mch_pyr(tmpdir_factory):
    out_im = tmpdir_factory.mktemp("image").join("image_fp_mch_pyr.tiff")
    subifds = 2
    full_im = np.random.randint(0, 255, (3, 2048, 2048), dtype=np.uint16)
    with TiffWriter(out_im) as tif:
        for ch in range(full_im.shape[0]):
            options = dict(
                tile=(256, 256),
                compression="deflate",
                photometric="minisblack",
                metadata=None,
            )
            tif.write(
                full_im[ch, :, :],
                subifds=subifds,
                **options,
            )

            for pyr_idx in range(subifds):
                if pyr_idx == 0:
                    subresimage = full_im[ch, ::2, ::2]
                else:
                    subresimage = subresimage[::2, ::2]

                tif.write(subresimage, **options, subfiletype=1)
    return out_im


@pytest.fixture
def disk_im_rgb_pyr(tmpdir_factory):
    out_im = tmpdir_factory.mktemp("image").join("image_fp_rgb_pyr.tiff")
    subifds = 2
    full_im = np.random.randint(0, 255, (2048, 2048, 3), dtype=np.uint8)
    with TiffWriter(out_im) as tif:
        options = dict(
            tile=(128, 128),
            compression="deflate",
            photometric="rgb",
            metadata=None,
        )
        tif.write(
            full_im,
            subifds=subifds,
            **options,
        )

        for pyr_idx in range(subifds):
            if pyr_idx == 0:
                subresimage = full_im[::2, ::2, :]
            else:
                subresimage = subresimage[::2, ::2, :]

            tif.write(subresimage, **options, subfiletype=1)

    return out_im


@pytest.fixture
def disk_im_gry_pyr(tmpdir_factory):
    out_im = tmpdir_factory.mktemp("image").join("image_fp_gry_pyr.tiff")
    subifds = 2
    full_im = np.random.randint(0, 255, (2048, 2048), dtype=np.uint16)
    with TiffWriter(out_im) as tif:
        options = dict(
            tile=(256, 256),
            compression="deflate",
            photometric="minisblack",
            metadata=None,
        )
        tif.write(
            full_im,
            subifds=subifds,
            **options,
        )

        for pyr_idx in range(subifds):
            if pyr_idx == 0:
                subresimage = full_im[::2, ::2]
            else:
                subresimage = subresimage[::2, ::2]

            tif.write(subresimage, **options, subfiletype=1)
    return out_im



from pathlib import Path
from typing import List, Optional, Tuple, Union

import cv2
import numpy as np
import SimpleITK as sitk
from tifffile import TiffWriter

from wsireg.reg_images import RegImage
from wsireg.reg_images.merge_reg_image import MergeRegImage
from wsireg.reg_transform_seq import RegTransformSeq
from wsireg.utils.im_utils import (
    SITK_TO_NP_DTYPE,
    format_channel_names,
    get_pyramid_info,
    prepare_ome_xml_str,
)


class MergeOmeTiffWriter:
    x_size: Optional[int] = None
    y_size: Optional[int] = None
    y_spacing: Optional[Union[int, float]] = None
    x_spacing: Optional[Union[int, float]] = None
    tile_size: int = 512
    pyr_levels: Optional[List[Tuple[int, int]]] = None
    n_pyr_levels: Optional[int] = None
    PhysicalSizeY: Optional[Union[int, float]] = None
    PhysicalSizeX: Optional[Union[int, float]] = None
    subifds: Optional[int] = None
    compression: str = "deflate"

    def __init__(
        self,
        reg_images: MergeRegImage,
        reg_seq_transforms: Optional[List[RegTransformSeq]] = None,
    ):
        self.reg_image = reg_images
        self.reg_seq_transforms = reg_seq_transforms

    def _length_checks(self, sub_image_names):

        if isinstance(sub_image_names, list) is False:
            if sub_image_names is None:
                sub_image_names = [
                    "" for i in range(len(self.reg_image.images))
                ]
            else:
                raise ValueError(
                    "MergeRegImage requires a list of image names for each image to merge"
                )

        if self.reg_seq_transforms is None:
            transformations = [None for i in range(len(self.reg_image.images))]
        else:
            transformations = self.reg_seq_transforms

        if len(transformations) != len(self.reg_image.images):
            raise ValueError(
                "MergeRegImage number of transforms does not match number of images"
            )

    def _create_channel_names(self, sub_image_names):
        def prepare_channel_names(sub_image_name, channel_names):
            return [f"{sub_image_name} - {c}" for c in channel_names]

        self.reg_image.channel_names = [
            prepare_channel_names(im_name, cnames)
            for im_name, cnames in zip(
                sub_image_names, self.reg_image.channel_names
            )
        ]
        self.reg_image.channel_names = [
            item
            for sublist in self.reg_image.channel_names
            for item in sublist
        ]

    def _transform_check(self):
        out_size = []
        out_spacing = []

        if not self.reg_seq_transforms:
            rts = [None for _ in range(len(self.reg_image.images))]
        else:
            rts = self.reg_seq_transforms
        for im, t in zip(self.reg_image.images, rts):
            if t:
                out_size.append(t.reg_transforms[-1].output_size)
                out_spacing.append(t.reg_transforms[-1].output_spacing)
            else:
                out_im_size = (
                    (im.im_dims[0], im.im_dims[1])
                    if im.is_rgb
                    else (im.im_dims[1], im.im_dims[2])
                )
                out_im_spacing = im.image_res

                out_size.append(out_im_size)
                out_spacing.append(out_im_spacing)

        if all(out_spacing) is False:
            raise ValueError(
                "MergeRegImage all transforms output spacings and untransformed image spacings must match"
            )

        if all(out_size) is False:
            raise ValueError(
                "MergeRegImage all transforms output sizes and untransformed image sizes must match"
            )

    def _prepare_image_info(
        self,
        reg_image: RegImage,
        image_name: str,
        im_dtype: np.dtype,
        reg_transform_seq: Optional[RegTransformSeq] = None,
        channel_names: Optional[List[str]] = None,
        write_pyramid: bool = True,
        tile_size: int = 512,
        compression: str = "default",
    ):

        if reg_transform_seq:
            self.x_size, self.y_size = reg_transform_seq.output_size
            self.x_spacing, self.y_spacing = reg_transform_seq.output_spacing
        else:
            self.y_size, self.x_size = (
                (reg_image.im_dims[0], reg_image.im_dims[1])
                if reg_image.is_rgb
                else (reg_image.im_dims[1], reg_image.im_dims[2])
            )
            self.y_spacing, self.x_spacing = (
                reg_image.image_res,
                reg_image.image_res,
            )

        self.tile_size = tile_size
        # protect against too large tile size
        while (
            self.y_size / self.tile_size <= 1
            or self.x_size / self.tile_size <= 1
        ):
            self.tile_size = self.tile_size // 2

        self.pyr_levels, _ = get_pyramid_info(
            self.y_size, self.x_size, reg_image.n_ch, self.tile_size
        )
        self.n_pyr_levels = len(self.pyr_levels)

        if reg_transform_seq:
            self.PhysicalSizeY = self.y_spacing
            self.PhysicalSizeX = self.x_spacing
        else:
            self.PhysicalSizeY = reg_image.image_res
            self.PhysicalSizeX = reg_image.image_res

        channel_names = format_channel_names(
            self.reg_image.channel_names, self.reg_image.n_ch
        )

        self.omexml = prepare_ome_xml_str(
            self.y_size,
            self.x_size,
            len(channel_names),
            im_dtype,
            False,
            PhysicalSizeX=self.PhysicalSizeX,
            PhysicalSizeY=self.PhysicalSizeY,
            PhysicalSizeXUnit="µm",
            PhysicalSizeYUnit="µm",
            Name=image_name,
            Channel={"Name": channel_names},
        )

        self.subifds = self.n_pyr_levels - 1 if write_pyramid is True else None

        if compression == "default":
            print("using default compression")
            self.compression = "deflate"
        else:
            self.compression = compression

    def _get_merge_dtype(self):
        dtype_max_size = [
            np.iinfo(r.im_dtype).max for r in self.reg_image.images
        ]

        merge_dtype_np = self.reg_image.images[
            np.argmax(dtype_max_size)
        ].im_dtype
        for k, v in SITK_TO_NP_DTYPE.items():
            if k < 12:
                if v == merge_dtype_np:
                    merge_dtype_sitk = k
        return merge_dtype_sitk, merge_dtype_np

    def merge_write_image_by_plane(
        self,
        image_name,
        sub_image_names,
        output_dir="",
        write_pyramid=True,
        tile_size=512,
        compression="default",
    ):
        merge_dtype_sitk, merge_dtype_np = self._get_merge_dtype()

        self._length_checks(sub_image_names)
        self._create_channel_names(sub_image_names)
        self._transform_check()

        output_file_name = str(Path(output_dir) / f"{image_name}.ome.tiff")

        self._prepare_image_info(
            self.reg_image.images[0],
            image_name,
            merge_dtype_np,
            reg_transform_seq=self.reg_seq_transforms[0],
            write_pyramid=write_pyramid,
            tile_size=tile_size,
            compression=compression,
        )

        print(f"saving to {output_file_name}")
        with TiffWriter(output_file_name, bigtiff=True) as tif:
            for m_idx, merge_image in enumerate(self.reg_image.images):
                if self.reg_image.images[m_idx].reader == "sitk":
                    full_image = sitk.ReadImage(
                        self.reg_image.images[m_idx].image_filepath
                    )
                merge_n_ch = merge_image.n_ch
                for channel_idx in range(merge_n_ch):
                    if self.reg_image.images[m_idx].reader != "sitk":
                        image = self.reg_image.images[
                            m_idx
                        ].read_single_channel(channel_idx)
                        image = np.squeeze(image)
                        image = sitk.GetImageFromArray(image)
                        image.SetSpacing(
                            (
                                self.reg_image.images[m_idx].image_res,
                                self.reg_image.images[m_idx].image_res,
                            )
                        )

                    else:
                        if len(full_image.GetSize()) > 2:
                            image = full_image[:, :, channel_idx]
                        else:
                            image = full_image

                    if image.GetPixelIDValue() != merge_dtype_sitk:
                        image = sitk.Cast(image, merge_dtype_sitk)

                    if self.reg_seq_transforms[m_idx]:
                        image = self.reg_seq_transforms[
                            m_idx
                        ].resampler.Execute(image)

                    if isinstance(image, sitk.Image):
                        image = sitk.GetArrayFromImage(image)

                    options = dict(
                        tile=(tile_size, tile_size),
                        compression=self.compression,
                        photometric="minisblack",
                        metadata=None,
                    )
                    # write OME-XML to the ImageDescription tag of the first page
                    description = (
                        self.omexml
                        if channel_idx == 0 and m_idx == 0
                        else None
                    )
                    # write channel data
                    print(
                        f" writing subimage index {m_idx} : {sub_image_names[m_idx]} - "
                        f"channel index - {channel_idx} - shape: {image.shape}"
                    )
                    tif.write(
                        image,
                        subifds=self.subifds,
                        description=description,
                        **options,
                    )

                    if write_pyramid:
                        for pyr_idx in range(1, self.n_pyr_levels):
                            resize_shape = (
                                self.pyr_levels[pyr_idx][0],
                                self.pyr_levels[pyr_idx][1],
                            )
                            image = cv2.resize(
                                image,
                                resize_shape,
                                cv2.INTER_LINEAR,
                            )

                            tif.write(image, **options, subfiletype=1)

            return output_file_name



from warnings import warn

import numpy as np
import SimpleITK as sitk

from wsireg.utils.tform_conversion import convert_to_itk


class RegTransform:
    def __init__(self, elastix_transform):
        self.elastix_transform = elastix_transform
        self.itk_transform = convert_to_itk(self.elastix_transform)

        self.output_spacing = [
            float(p) for p in self.elastix_transform.get("Spacing")
        ]
        self.output_size = [int(p) for p in self.elastix_transform.get("Size")]
        self.output_origin = [
            float(p) for p in self.elastix_transform.get("Origin")
        ]
        self.output_direction = [
            float(p) for p in self.elastix_transform.get("Direction")
        ]
        self.resample_interpolator = self.elastix_transform.get(
            "ResampleInterpolator"
        )[0]

        self.is_linear = self.itk_transform.IsLinear()

        if self.is_linear is True:
            self.inverse_transform = self.itk_transform.GetInverse()
            transform_name = self.itk_transform.GetName()
            if transform_name == "Euler2DTransform":
                self.inverse_transform = sitk.Euler2DTransform(
                    self.inverse_transform
                )
            elif transform_name == "AffineTransform":
                self.inverse_transform = sitk.AffineTransform(
                    self.inverse_transform
                )
            elif transform_name == "Similarity2DTransform":
                self.inverse_transform = sitk.Similarity2DTransform(
                    self.inverse_transform
                )
        else:
            self.inverse_transform = None

    def compute_inverse_nonlinear(self):

        tform_to_dfield = sitk.TransformToDisplacementFieldFilter()
        tform_to_dfield.SetOutputSpacing(self.output_spacing)
        tform_to_dfield.SetOutputOrigin(self.output_origin)
        tform_to_dfield.SetOutputDirection(self.output_direction)
        tform_to_dfield.SetSize(self.output_size)

        displacement_field = tform_to_dfield.Execute(self.itk_transform)
        displacement_field = sitk.InvertDisplacementField(displacement_field)
        displacement_field = sitk.DisplacementFieldTransform(
            displacement_field
        )

        self.inverse_transform = displacement_field

    def as_np_matrix(
        self,
        use_np_ordering=False,
        n_dim=3,
        use_inverse=False,
        to_px_idx=False,
    ):
        if self.is_linear:
            if use_np_ordering is True:
                order = slice(None, None, -1)
            else:
                order = slice(None, None, 1)

            if use_inverse is True:
                transform = self.inverse_transform
            else:
                transform = self.itk_transform

            # pull transform values
            tmatrix = np.array(transform.GetMatrix()[order]).reshape(2, 2)
            center = np.array(transform.GetCenter()[order])
            translation = np.array(transform.GetTranslation()[order])

            if to_px_idx is True:
                phys_to_index = 1 / np.asarray(self.output_spacing).astype(
                    np.float64
                )
                center *= phys_to_index
                translation *= phys_to_index

            # construct matrix
            full_matrix = np.eye(n_dim)
            full_matrix[0:2, 0:2] = tmatrix
            full_matrix[0:2, n_dim - 1] = (
                -np.dot(tmatrix, center) + translation + center
            )

            return full_matrix
        else:
            warn(
                "Non-linear transformations can not be represented converted"
                "to homogenous matrix"
            )
            return None



import pytest

from tests.fixtures.im_fixtures import (
    dask_im_gry_np,
    dask_im_mch_np,
    dask_im_rgb_np,
    disk_im_gry,
    disk_im_gry_pyr,
    disk_im_mch,
    disk_im_mch_notile,
    disk_im_mch_pyr,
    disk_im_rgb,
    disk_im_rgb_pyr,
    im_gry_np,
    im_mch_np,
    im_rgb_np,
    mask_np,
    zarr_im_gry_np,
    zarr_im_mch_np,
    zarr_im_rgb_np,
)
from tests.fixtures.transform_fixtures import (
    complex_transform,
    complex_transform_larger,
    complex_transform_larger_padded,
    simple_transform_affine,
    simple_transform_affine_large_output,
    simple_transform_affine_nl,
    simple_transform_affine_nl_large_output,
)



import json
import time
from copy import copy, deepcopy
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union
from warnings import warn

import numpy as np
import SimpleITK as sitk
import yaml

from wsireg.parameter_maps.preprocessing import ImagePreproParams
from wsireg.parameter_maps.reg_model import RegModel
from wsireg.reg_images import RegImage, MergeRegImage
from wsireg.reg_images.loader import reg_image_loader
from wsireg.reg_shapes import RegShapes
from wsireg.reg_transform import RegTransform
from wsireg.reg_transform_seq import RegTransformSeq
from wsireg.utils.config_utils import parse_check_reg_config
from wsireg.utils.im_utils import ARRAYLIKE_CLASSES
from wsireg.utils.reg_utils import (
    _prepare_reg_models,
    json_to_pmap_dict,
    pmap_dict_to_json,
    register_2d_images_itkelx,
    sitk_pmap_to_dict,
)
from wsireg.utils.shape_utils import invert_nonrigid_transforms
from wsireg.utils.tform_utils import identity_elx_transform
from wsireg.writers.merge_ome_tiff_writer import MergeOmeTiffWriter
from wsireg.writers.ome_tiff_writer import OmeTiffWriter


class WsiReg2D(object):
    project_name: Optional[str] = None
    output_dir: Optional[Union[str, Path]] = None
    image_cache: Optional[Path] = None
    """
    Class to define a 2D registration graph and execute the registrations and transformations of the graph

    Parameters
    ----------
    project_name: str
        Project name will prefix all output files and directories
    output_dir: str
        Directory where registration data will be stored
    cache_images: bool
        whether to store images as they are preprocessed for registration (if you need to repeat or modify settings
        this will avoid image io and preprocessing)
    config: str or Path
        path to a 2D wsireg YAML configuration

    Attributes
    ----------
    modalities: dict
        dictionary of modality information (file path, spatial res., preprocessing), defines a graph node

    modalities: list
        list of all modality names

    n_modalities: int
        number of modalities (nodes) in the graphs

    reg_paths: dict
        dictionary of a modalities path from node to node

    reg_graph_edges: dict
        generated dictionary of necessary registrations to move modalities to their target modality

    n_registrations: int
        number of explicit registrations (edges) in the graphs

    transformation_paths: dict
        generated dictionary of necessary source - target transformations to transform modalities to their target modality

    transformations: dict
        per modality dictionary containing transformation parameters for each registration

    attachment_images: dict
        images to be transformed along the path of the defined graph, assoicated to a given modality (masks, other registered images)

    attachment_shapes: dict
        shape data attached to a modality to be transformed along the graph


    """

    def __init__(
        self,
        project_name: Optional[str] = None,
        output_dir: Optional[Union[str, Path]] = None,
        cache_images: bool = True,
        config: Optional[Union[str, Path]] = None,
    ):

        self.setup_project_output(project_name, output_dir)

        self.cache_images = cache_images

        self.pairwise = False

        self._modalities = dict()
        self._modality_names = []
        self._reg_paths = dict()
        self._reg_graph_edges = []
        self._transform_paths = dict()

        self._transformations = None
        self._preprocessed_image_sizes: Dict[str, Tuple[int, int]] = dict()
        self._preprocessed_image_spacings: Dict[
            str, Union[Tuple[int, int], Tuple[float, float]]
        ] = dict()

        self.n_modalities: Optional[int] = None
        self.n_registrations: Optional[int] = None

        self.attachment_images = dict()

        self._shape_sets = dict()
        self._shape_set_names = []

        self.merge_modalities = dict()
        self.original_size_transforms = dict()

        if config:
            self.add_data_from_config(config)

    def setup_project_output(
        self,
        project_name: Optional[str] = None,
        output_dir: Optional[Union[str, Path]] = None,
    ) -> None:
        if project_name is None:
            self.project_name = 'RegProj'
        else:
            self.project_name = project_name

        if output_dir is None:
            output_dir = "./"

        self.output_dir = Path(output_dir)

        self.image_cache = self.output_dir / ".imcache_{}".format(project_name)

    @property
    def modalities(self):
        return self._modalities

    @modalities.setter
    def modalities(self, modality):
        self._modalities.update(modality)
        self.n_modalities = len(self._modalities)

    @property
    def shape_sets(self):
        return self._shape_sets

    @shape_sets.setter
    def shape_sets(self, shape_set):
        self._shape_sets.update(shape_set)

    @property
    def shape_set_names(self):
        return self._shape_set_names

    @shape_set_names.setter
    def shape_set_names(self, shape_set_name):
        self._shape_set_names.append(shape_set_name)

    @property
    def modality_names(self):
        return self._modality_names

    @modality_names.setter
    def modality_names(self, modality_name):
        self._modality_names.append(modality_name)

    def add_modality(
        self,
        modality_name: str,
        image_fp: Union[Path, str, np.ndarray],
        image_res: Union[int, float] = 1,
        channel_names: Optional[List[str]] = None,
        channel_colors: Optional[List[str]] = None,
        preprocessing: Optional[
            Union[ImagePreproParams, Dict[str, Any]]
        ] = None,
        mask: Optional[Union[str, Path, np.ndarray]] = None,
        prepro_dict: Optional[Dict[str, Any]] = None,
        output_res: Optional[
            Union[int, float, Tuple[int, int], Tuple[float, float]]
        ] = None,
    ) -> None:
        """
        Add an image modality (node) to the registration graph

        Parameters
        ----------
        modality_name : str
            Unique name identifier for the modality
        image_fp : str
            file path to the image to be read
        image_res : float
            spatial resolution of image in units per px (i.e. 0.9 um / px)
        channel_names: List[str]
            names for the channels to go into the OME-TIFF
        channel_colors: List[str]
            channels colors for OME-TIFF (not implemented)
        mask: Union[str, Path, np.ndarray]
            path to binary mask (>0 is in) image for registration and/or cropping
        preprocessing :
            preprocessing parameters for the modality for registration. Registration images should be a xy single plane
            so many modalities (multi-channel, RGB) must "create" a single channel.
            Defaults: multi-channel images -> max intensity project image
            RGB -> greyscale then intensity inversion (black background, white foreground)
        output_res : Union[Tuple[int,int], Tuple[float,float]]
            change output spacing/resolution when resampling images, default will be the spacing
            of the final target image
        """
        if modality_name in self._modality_names:
            raise ValueError(
                'modality named \"{}\" is already in modality_names'.format(
                    modality_name
                )
            )
        if prepro_dict:
            preprocessing = prepro_dict
            DeprecationWarning(
                "'prepro_dict' argument is deprepcated and will be removed\
                in a future release, use 'preprocessing' instead"
            )
        if preprocessing:
            if isinstance(preprocessing, dict):
                image_prepro = ImagePreproParams(**preprocessing)
            else:
                image_prepro = preprocessing
        else:
            image_prepro = ImagePreproParams()

        if isinstance(output_res, (int, float)):
            output_res = (output_res, output_res)

        self.modalities = {
            modality_name: {
                "image_filepath": image_fp,
                "image_res": image_res,
                "channel_names": channel_names,
                "channel_colors": channel_colors,
                "preprocessing": image_prepro,
                "mask": mask,
                "output_res": output_res,
            }
        }

        self.modality_names = modality_name

    def add_shape_set(
        self,
        attachment_modality: str,
        shape_set_name: str,
        shape_files: List[Union[str, Path]],
        image_res: Union[int, float],
    ) -> None:
        """
        Add a shape set to the graph

        Parameters
        ----------
        attachment_modality : str
            image modality to which the shapes are attached
        shape_set_name : str
            Unique name identifier for the shape set
        shape_files : list of file paths
            list of shape data in geoJSON format or list of dicts containing following keys:
            "array" = np.ndarray of xy coordinates, "shape_type" = geojson shape type (Polygon, MultiPolygon, etc.),
            "shape_name" = class of the shape("tumor","normal",etc.)
        image_res : float
            spatial resolution of shape data's associated image in units per px (i.e. 0.9 um / px)
        """
        if shape_set_name in self._shape_set_names:
            raise ValueError(
                'shape named \"{}\" is already in shape_set_names'.format(
                    shape_set_name
                )
            )

        self.shape_sets = {
            shape_set_name: {
                "shape_files": shape_files,
                "image_res": image_res,
                "attachment_modality": attachment_modality,
            }
        }

        self.shape_set_names = shape_set_name

    def add_attachment_images(
        self,
        attachment_modality: str,
        modality_name: str,
        image_fp: Union[str, Path],
        image_res: Union[int, float] = 1,
        channel_names: Optional[List[str]] = None,
        channel_colors: Optional[List[str]] = None,
    ) -> None:
        """
        Images which are unregistered between modalities, but are transformed following the path of one of the graph's
        modalities.

        Parameters
        ----------
        attachment_modality : str
            image modality to which the new image are attached
        modality_name : str
            name of the added attachment image
        image_fp : str
            path to the attachment modality, it will be imported and transformed without preprocessing
        image_res : float
            spatial resolution of attachment image data's in units per px (i.e. 0.9 um / px)
        channel_names: List[str]
            names for the channels to go into the OME-TIFF
        channel_colors: List[str]
            channels colors for OME-TIFF (not implemented)
        """
        if attachment_modality not in self.modality_names:
            raise ValueError(
                'attachment modality named \"{}\" not found in modality_names'.format(
                    attachment_modality
                )
            )
        self.add_modality(
            modality_name,
            image_fp,
            image_res,
            channel_names=channel_names,
            channel_colors=channel_colors,
        )
        self.attachment_images[modality_name] = attachment_modality

    def add_attachment_shapes(
        self,
        attachment_modality: str,
        shape_set_name: str,
        shape_files: List[Union[str, Path]],
    ) -> None:
        """
        Add attached shapes

        Parameters
        ----------
        attachment_modality : str
            image modality to which the shapes are attached
        shape_set_name : str
            Unique name identifier for the shape set
        shape_files : list of file paths
            list of shape data in geoJSON format or list of dicts containing following keys:
            "array" = np.ndarray of xy coordinates, "shape_type" = geojson shape type (Polygon, MultiPolygon, etc.),
            "shape_name" = class of the shape("tumor","normal",etc.)
        """
        if attachment_modality not in self.modality_names:
            raise ValueError(
                'attachment modality \"{}\" for shapes \"{}\" not found in modality_names {}'.format(
                    attachment_modality, shape_set_name, self.modality_names
                )
            )

        image_res = self.modalities[attachment_modality]["image_res"]
        self.add_shape_set(
            attachment_modality, shape_set_name, shape_files, image_res
        )

    @property
    def reg_paths(self):
        return self._reg_paths

    @reg_paths.setter
    def reg_paths(self, path_values):
        (
            src_modality,
            tgt_modality,
            thru_modality,
            reg_params,
            override_prepro,
        ) = path_values

        if thru_modality != tgt_modality:
            self._reg_paths.update(
                {src_modality: [thru_modality, tgt_modality]}
            )
        else:
            self._reg_paths.update({src_modality: [tgt_modality]})

        self.reg_graph_edges = {
            'modalities': {'source': src_modality, 'target': thru_modality},
            'params': reg_params,
            "override_prepro": override_prepro,
        }
        self.transform_paths = self._reg_paths

    def add_reg_path(
        self,
        src_modality_name: str,
        tgt_modality_name: str,
        thru_modality: Optional[str] = None,
        reg_params: Union[str, RegModel, List[str], List[RegModel]] = [
            RegModel.rigid
        ],
        override_prepro: dict = {"source": None, "target": None},
    ):
        """
        Add registration path between modalities as well as a thru modality that describes where to attach edges.

        Parameters
        ----------
        src_modality_name : str
            modality that has been added to graph to be transformed to tgt_modality
        tgt_modality_name : str
            modality that has been added to graph that is being aligned to
        thru_modality: str
            modality that has been added to graph by which another should be run through
        reg_params:
            SimpleElastix registration parameters
        override_prepro:
            set specific preprocessing for a given registration edge for the source or target image that will override
            the set modality preprocessing FOR THIS REGISTRATION ONLY.
        """
        if src_modality_name not in self.modality_names:
            raise ValueError("source modality not found!")
        if tgt_modality_name not in self.modality_names:
            raise ValueError("target modality not found!")

        if thru_modality is None:
            self.reg_paths = (
                src_modality_name,
                tgt_modality_name,
                tgt_modality_name,
                reg_params,
                override_prepro,
            )
        else:
            self.reg_paths = (
                src_modality_name,
                tgt_modality_name,
                thru_modality,
                reg_params,
                override_prepro,
            )

    @property
    def reg_graph_edges(self):
        return self._reg_graph_edges

    @reg_graph_edges.setter
    def reg_graph_edges(self, edge):
        self._reg_graph_edges.append(edge)
        self.n_registrations = len(self._reg_graph_edges)

    @property
    def transform_paths(self):
        return self._transform_paths

    @transform_paths.setter
    def transform_paths(self, reg_paths):

        transform_path_dict = {}

        for k, v in reg_paths.items():
            tform_path = self.find_path(k, v[-1])
            if self.pairwise is True:
                tform_path_modalities = tform_path[:1]
            else:
                tform_path_modalities = tform_path
            transform_edges = []
            for modality in tform_path_modalities:
                for edge in self.reg_graph_edges:
                    edge_modality = edge["modalities"]['source']
                    if modality == edge_modality:
                        transform_edges.append(edge["modalities"])
            transform_path_dict.update({k: transform_edges})

        self._transform_paths = transform_path_dict

    def find_path(
        self,
        start_modality: str,
        end_modality: str,
        path: Optional[List[str]] = None,
    ) -> Optional[List[str]]:
        """
        Find a path from start_modality to end_modality in the graph
        """
        if path is None:
            path = []
        path = path + [start_modality]
        if start_modality == end_modality:
            return path
        if start_modality not in self.reg_paths:
            return None
        for modality in self.reg_paths[start_modality]:
            if modality not in path:
                extended_path = self.find_path(modality, end_modality, path)
                if extended_path:
                    return extended_path
        return None

    def _check_cache_modality(self, modality_name: str):
        cache_im_fp = self.image_cache / "{}_prepro.tiff".format(modality_name)
        cache_transform_fp = cache_im_fp.parent / "{}_init_tforms.json".format(
            cache_im_fp.stem
        )
        cache_osize_tform_fp = (
            self.image_cache
            / "{}_orig_size_tform.json".format(cache_im_fp.stem)
        )

        if cache_im_fp.exists() is True:
            im_fp = str(cache_im_fp)
            im_from_cache = True
        else:
            im_fp = self.modalities[modality_name]["image_filepath"]
            im_from_cache = False

        if cache_transform_fp.exists() is True:
            im_initial_transforms = [json_to_pmap_dict(cache_transform_fp)]
        else:
            im_initial_transforms = None

        if cache_osize_tform_fp.exists() is True:
            osize_tform = json_to_pmap_dict(cache_osize_tform_fp)
        else:
            osize_tform = None

        return im_fp, im_initial_transforms, im_from_cache, osize_tform

    def _prepare_modality(
        self, modality_name: str, reg_edge: dict, src_or_tgt: str
    ) -> Tuple:
        mod_data = self.modalities[modality_name].copy()

        if reg_edge.get("override_prepro") is not None:
            override_preprocessing = reg_edge.get("override_prepro")[
                src_or_tgt
            ]
        else:
            override_preprocessing = None

        if override_preprocessing is not None:
            mod_data["preprocessing"] = override_preprocessing

            return (
                mod_data["image_filepath"],
                mod_data["image_res"],
                mod_data["preprocessing"],
                None,
                mod_data["mask"],
                None,
            )
        else:

            (
                mod_data["image_filepath"],
                mod_data["transforms"],
                im_from_cache,
                original_size_transform,
            ) = self._check_cache_modality(modality_name)

            if im_from_cache:
                if mod_data["preprocessing"].use_mask is False:
                    mod_data["mask"] = None
                if mod_data["preprocessing"].downsampling > 1:
                    image_res = (
                        mod_data["image_res"]
                        * mod_data["preprocessing"].downsampling
                    )
                else:
                    image_res = mod_data["image_res"]
                mod_data["preprocessing"] = None
            else:
                image_res = mod_data["image_res"]

            return (
                mod_data["image_filepath"],
                image_res,
                mod_data["preprocessing"],
                mod_data["transforms"],
                mod_data["mask"],
                original_size_transform,
            )

    def _cache_images(self, modality_name: str, reg_image: RegImage) -> None:

        cache_im_fp = self.image_cache / "{}_prepro.tiff".format(modality_name)

        cache_transform_fp = self.image_cache / "{}_init_tforms.json".format(
            cache_im_fp.stem
        )

        cache_osize_tform_fp = (
            self.image_cache
            / "{}_orig_size_tform.json".format(cache_im_fp.stem)
        )
        if cache_im_fp.is_file() is False:
            sitk.WriteImage(reg_image.reg_image, str(cache_im_fp), True)

        if reg_image.mask is not None:
            cache_mask_im_fp = self.image_cache / "{}_prepro_mask.tiff".format(
                modality_name
            )
            if cache_mask_im_fp.is_file() is False:
                sitk.WriteImage(reg_image.mask, str(cache_mask_im_fp), True)

        if cache_transform_fp.is_file() is False:
            pmap_dict_to_json(
                reg_image.pre_reg_transforms, str(cache_transform_fp)
            )

        if (
            cache_osize_tform_fp.is_file() is False
            and reg_image.original_size_transform is not None
        ):
            pmap_dict_to_json(
                reg_image.original_size_transform, str(cache_osize_tform_fp)
            )

    def _find_nonreg_modalities(self) -> None:
        registered_modalities = [
            edge.get("modalities").get("source")
            for edge in self.reg_graph_edges
        ]
        non_reg_modalities = list(
            set(self.modality_names).difference(registered_modalities)
        )

        # remove attachment modalities
        for attachment_modality in self.attachment_images.keys():
            non_reg_modalities.pop(
                non_reg_modalities.index(attachment_modality)
            )

        return non_reg_modalities

    def save_config(self, registered=False):
        ts = time.strftime('%Y%m%d-%H%M%S')
        status = "registered" if registered is True else "setup"

        reg_paths = {}
        for idx, edge in enumerate(self.reg_graph_edges):
            src_modality = edge.get("modalities").get("source")
            if len(self.reg_paths[src_modality]) > 1:
                thru_modality = self.reg_paths[src_modality][0]
            else:
                thru_modality = None
            tgt_modality = self.reg_paths[src_modality][-1]
            reg_paths.update(
                {
                    f"reg_path_{idx}": {
                        "src_modality_name": edge.get("modalities").get(
                            "source"
                        ),
                        "tgt_modality_name": tgt_modality,
                        "thru_modality": thru_modality,
                        "reg_params": edge.get("params"),
                    }
                }
            )

        reg_graph_edges = deepcopy(self.reg_graph_edges)

        [rge.pop("transforms", None) for rge in reg_graph_edges]

        modalities_out = deepcopy(self.modalities)
        for mod, data in modalities_out.items():
            if isinstance(data["image_filepath"], ARRAYLIKE_CLASSES):
                data["image_filepath"] = "ArrayLike"

        config = {
            "project_name": self.project_name,
            "output_dir": str(self.output_dir),
            "cache_images": self.cache_images,
            "modalities": modalities_out,
            "reg_paths": reg_paths,
            "reg_graph_edges": reg_graph_edges
            if status == "registered"
            else None,
            "original_size_transforms": self.original_size_transforms
            if status == "registered"
            else None,
            "attachment_shapes": self.shape_sets
            if len(self._shape_sets) > 0
            else None,
            "attachment_images": self.attachment_images
            if len(self.attachment_images) > 0
            else None,
        }

        output_path = (
            self.output_dir
            / f"{ts}-{self.project_name}-configuration-{status}.yaml"
        )

        with open(str(output_path), "w") as f:
            yaml.dump(config, f, sort_keys=False)

    def register_images(self, parallel=False):
        """
        Start image registration process for all modalities

        Parameters
        ----------
        parallel : bool
            whether to run each edge in parallel (not implemented yet)
        """
        if self.cache_images is True:
            self.image_cache.mkdir(parents=False, exist_ok=True)

        self.save_config(registered=False)

        for reg_edge in self.reg_graph_edges:
            if (
                reg_edge.get("registered") is None
                or reg_edge.get("registered") is False
            ):
                src_name = reg_edge["modalities"]["source"]
                tgt_name = reg_edge["modalities"]["target"]

                (
                    src_reg_image_fp,
                    src_res,
                    src_prepro,
                    src_transforms,
                    src_mask,
                    src_original_size_transform,
                ) = self._prepare_modality(src_name, reg_edge, "source")

                (
                    tgt_reg_image_fp,
                    tgt_res,
                    tgt_prepro,
                    tgt_transforms,
                    tgt_mask,
                    tgt_original_size_transform,
                ) = self._prepare_modality(tgt_name, reg_edge, "target")

                src_reg_image = reg_image_loader(
                    src_reg_image_fp,
                    src_res,
                    preprocessing=src_prepro,
                    pre_reg_transforms=src_transforms,
                    mask=src_mask,
                )

                tgt_reg_image = reg_image_loader(
                    tgt_reg_image_fp,
                    tgt_res,
                    preprocessing=tgt_prepro,
                    pre_reg_transforms=tgt_transforms,
                    mask=tgt_mask,
                )

                src_reg_image.read_reg_image()
                tgt_reg_image.read_reg_image()

                self._preprocessed_image_spacings.update(
                    {src_name: src_reg_image.reg_image.GetSpacing()}
                )

                self._preprocessed_image_spacings.update(
                    {tgt_name: tgt_reg_image.reg_image.GetSpacing()}
                )
                self._preprocessed_image_sizes.update(
                    {src_name: src_reg_image.reg_image.GetSize()}
                )
                self._preprocessed_image_sizes.update(
                    {tgt_name: tgt_reg_image.reg_image.GetSize()}
                )

                if (
                    tgt_original_size_transform is None
                    and tgt_reg_image.original_size_transform is not None
                ):
                    tgt_original_size_transform = (
                        tgt_reg_image.original_size_transform
                    )

                if self.cache_images is True:
                    if reg_edge.get("override_prepro") is not None:
                        if (
                            reg_edge.get("override_prepro").get("source")
                            is None
                        ):
                            self._cache_images(src_name, src_reg_image)
                        if (
                            reg_edge.get("override_prepro").get("target")
                            is None
                        ):
                            self._cache_images(tgt_name, tgt_reg_image)
                    else:
                        self._cache_images(src_name, src_reg_image)
                        self._cache_images(tgt_name, tgt_reg_image)

                reg_params = reg_edge["params"]

                output_path = (
                    self.output_dir
                    / "{}-{}_to_{}_reg_output".format(
                        self.project_name,
                        reg_edge["modalities"]["source"],
                        reg_edge["modalities"]["target"],
                    )
                )

                output_path.mkdir(parents=False, exist_ok=True)

                reg_params_prepared = _prepare_reg_models(reg_params)

                reg_tforms = register_2d_images_itkelx(
                    src_reg_image,
                    tgt_reg_image,
                    reg_params_prepared,
                    output_path,
                )

                reg_tforms = [sitk_pmap_to_dict(tf) for tf in reg_tforms]

                if src_transforms is not None:
                    initial_transforms = src_transforms[0]
                else:
                    initial_transforms = src_reg_image.pre_reg_transforms

                if initial_transforms:
                    initial_transforms_rt = [
                        RegTransform(t) for t in initial_transforms
                    ]
                    initial_transforms_idx = [
                        idx for idx, _ in enumerate(initial_transforms_rt)
                    ]
                    initial_rt_seq = RegTransformSeq(
                        initial_transforms_rt, initial_transforms_idx
                    )

                reg_tforms_rt = [RegTransform(t) for t in reg_tforms]
                reg_tforms_idx = [0 for _ in reg_tforms_rt]
                reg_rt_seq = RegTransformSeq(reg_tforms_rt, reg_tforms_idx)

                reg_edge["transforms"] = {
                    'initial': initial_rt_seq if initial_transforms else None,
                    'registration': reg_rt_seq,
                }

                self.original_size_transforms.update(
                    {tgt_name: tgt_original_size_transform}
                )

                reg_edge["registered"] = True

        self.transformations = self.reg_graph_edges
        # self.save_config(registered=True)

    @property
    def transformations(self):
        return self._transformations

    @transformations.setter
    def transformations(self, reg_graph_edges):
        self._transformations = self._collate_transformations()

    def add_merge_modalities(self, merge_name, modalities):
        for modality in modalities:
            try:
                self.modalities[modality]
            except KeyError:
                raise ValueError(
                    f"Modality for merger [{modality}] is not a modality "
                    f"within the graph, current modalitles : "
                    f"{self.modality_names}"
                )
            self.merge_modalities.update({merge_name: modalities})

    def _generate_reg_transforms(self):
        self._reg_graph_edges["reg_transforms"]

    def _collate_transformations(self):
        transforms = dict()
        edge_modality_pairs = [v['modalities'] for v in self.reg_graph_edges]
        for modality, tform_edges in self.transform_paths.items():
            full_tform_seq = RegTransformSeq()
            for idx, tform_edge in enumerate(tform_edges):
                reg_edge_tforms = self.reg_graph_edges[
                    edge_modality_pairs.index(tform_edge)
                ]["transforms"]
                if idx == 0:
                    transforms[modality] = {
                        f"initial-{tform_edges[idx]['source']}": reg_edge_tforms[
                            'initial'
                        ],
                        f"{str(idx).zfill(3)}-to-{tform_edges[idx]['target']}": reg_edge_tforms[
                            'registration'
                        ],
                    }
                    if reg_edge_tforms['initial']:
                        full_tform_seq.append(reg_edge_tforms['initial'])
                    full_tform_seq.append(reg_edge_tforms['registration'])
                else:
                    transforms[modality][
                        f"{str(idx).zfill(3)}-to-{tform_edges[idx]['target']}"
                    ] = reg_edge_tforms['registration']
                    full_tform_seq.append(reg_edge_tforms['registration'])

                transforms[modality]["full-transform-seq"] = full_tform_seq

        return transforms

    def _prepare_nonreg_image_transform(
        self,
        modality_key,
        attachment=False,
        attachment_modality=None,
        to_original_size=True,
    ):

        print(
            "preparing transforms for non-registered modality : {} ".format(
                modality_key
            )
        )
        output_path = self.output_dir / "{}-{}_registered".format(
            self.project_name, modality_key
        )

        if attachment:
            im_data_key = copy(modality_key)
            modality_key = attachment_modality

        transformations = None

        im_data = self.modalities[modality_key]

        if (
            im_data["preprocessing"].rot_cc != 0
            or im_data["preprocessing"].flip
            or im_data["preprocessing"].crop_to_mask_bbox
            or im_data["preprocessing"].mask_bbox
        ):
            _, im_initial_transforms, _, _ = self._check_cache_modality(
                modality_key
            )

            transformations = RegTransformSeq(
                [RegTransform(t) for t in im_initial_transforms[0]],
                transform_seq_idx=[
                    idx for idx in range(len(im_initial_transforms))
                ],
            )

        if to_original_size and self.original_size_transforms[modality_key]:
            orig_size_rt = RegTransformSeq(
                RegTransform(self.original_size_transforms[modality_key]),
                transform_seq_idx=[0],
            )
            if transformations:
                transformations.append(orig_size_rt)
            else:
                transformations = orig_size_rt

        if im_data["preprocessing"].downsampling > 1 and transformations:
            if not im_data["output_res"]:
                output_spacing_target = im_data["image_res"]
                transformations.set_output_spacing(
                    (output_spacing_target, output_spacing_target)
                )
            else:
                transformations.set_output_spacing(im_data["output_res"])

        elif im_data["preprocessing"].downsampling > 1 and not transformations:
            transformations = RegTransformSeq(
                [
                    RegTransform(
                        identity_elx_transform(
                            self._preprocessed_image_sizes[modality_key],
                            self._preprocessed_image_spacings[modality_key],
                        )
                    )
                ],
                transform_seq_idx=[0],
            )

            if not im_data["output_res"]:
                output_spacing_target = im_data["image_res"]
                transformations.set_output_spacing(
                    (output_spacing_target, output_spacing_target)
                )
            else:
                transformations.set_output_spacing(im_data["output_res"])

        if attachment:
            im_data = self.modalities[im_data_key]

        return im_data, transformations, output_path

    def _prepare_reg_image_transform(
        self,
        edge_key,
        attachment=False,
        attachment_modality=None,
        to_original_size=True,
    ):
        if attachment:
            final_modality = self.reg_paths[attachment_modality][-1]
            transformations = copy(
                self.transformations[attachment_modality]["full-transform-seq"]
            )
        else:
            final_modality = self.reg_paths[edge_key][-1]
            transformations = copy(
                self.transformations[edge_key]["full-transform-seq"]
            )

        output_path = self.output_dir / "{}-{}_to_{}_registered".format(
            self.project_name,
            edge_key,
            final_modality,
        )

        if attachment:
            im_data_key = copy(edge_key)
            edge_key = attachment_modality

        im_data = self.modalities[edge_key]

        if (
            self.original_size_transforms.get(final_modality)
            and to_original_size
        ):
            print("adding transform to original size")
            original_size_transform = self.original_size_transforms[
                final_modality
            ]
            orig_size_rt = RegTransformSeq(
                RegTransform(original_size_transform), transform_seq_idx=[0]
            )
            transformations.append(orig_size_rt)

        if im_data["preprocessing"].downsampling > 1:
            if not im_data["output_res"]:
                output_spacing_target = self.modalities[final_modality][
                    "image_res"
                ]
                transformations.set_output_spacing(
                    (output_spacing_target, output_spacing_target)
                )
            else:
                transformations.set_output_spacing(im_data["output_res"])

        elif im_data["output_res"]:
            transformations.set_output_spacing(im_data["output_res"])

        if attachment:
            im_data = self.modalities[im_data_key]

        return im_data, transformations, output_path

    def _transform_write_image(
        self, im_data, transformations, output_path, file_writer="ome.tiff"
    ):

        tfregimage = reg_image_loader(
            im_data["image_filepath"],
            im_data["image_res"],
            channel_names=im_data.get("channel_names"),
            channel_colors=im_data.get("channel_colors"),
        )

        ometiffwriter = OmeTiffWriter(
            tfregimage, reg_transform_seq=transformations
        )

        if (
            file_writer == "ome.tiff-bytile"
            and ometiffwriter.reg_image.reader not in ["czi", "sitk"]
        ):
            im_fp = ometiffwriter.write_image_by_tile(
                output_path.stem,
                output_dir=str(self.output_dir),
            )
        else:
            im_fp = ometiffwriter.write_image_by_plane(
                output_path.stem,
                output_dir=str(self.output_dir),
            )

        return im_fp

    def _transform_write_merge_images(self, to_original_size=True):
        def determine_attachment(sub_image):
            if sub_image in self.attachment_images.keys():
                attachment_modality = self.attachment_images[sub_image]
                attachment = True
            else:
                attachment_modality = None
                attachment = False
            return attachment, attachment_modality

        for merge_name, sub_images in self.merge_modalities.items():
            im_fps = []
            im_res = []
            im_ch_names = []
            transformations = []

            non_reg_modalities = self._find_nonreg_modalities()

            for sub_image in sub_images:
                attachment, attachment_modality = determine_attachment(
                    sub_image
                )
                im_data = self.modalities[sub_image]
                im_fps.append(im_data["image_filepath"])
                im_res.append(im_data["image_res"])
                im_ch_names.append(im_data.get("channel_names"))
                if (
                    sub_image not in non_reg_modalities
                    and attachment_modality not in non_reg_modalities
                ):
                    (
                        _,
                        sub_im_transforms,
                        _,
                    ) = self._prepare_reg_image_transform(
                        sub_image,
                        attachment=attachment,
                        attachment_modality=attachment_modality,
                        to_original_size=to_original_size,
                    )

                else:

                    (
                        _,
                        sub_im_transforms,
                        _,
                    ) = self._prepare_nonreg_image_transform(
                        sub_image,
                        attachment=attachment,
                        attachment_modality=attachment_modality,
                        to_original_size=to_original_size,
                    )

                transformations.append(sub_im_transforms)

        output_path = self.output_dir / "{}-{}_merged-registered".format(
            self.project_name,
            merge_name,
        )

        merge_regimage = MergeRegImage(
            im_fps,
            im_res,
            channel_names=im_ch_names,
        )

        merge_ometiffwriter = MergeOmeTiffWriter(
            merge_regimage, reg_seq_transforms=transformations
        )

        im_fp = merge_ometiffwriter.merge_write_image_by_plane(
            output_path.stem,
            sub_images,
            output_dir=str(self.output_dir),
        )
        return im_fp

    def transform_images(
        self,
        file_writer="ome.tiff",
        transform_non_reg=True,
        remove_merged=True,
        to_original_size=True,
    ):
        """
        Transform and write images to disk after registration. Also transforms all attachment images

        Parameters
        ----------
        file_writer : str
            output type to use, sitk writes a single resolution tiff, "zarr" writes an ome-zarr multiscale
            zarr store
        transform_non_reg : bool
            whether to write the images that aren't transformed during registration as well
        remove_merged: bool
            whether to remove images that are stored in merged store, if True, images that are merged
            will not be written as individual images as well
        to_original_size: bool
            write images that have been cropped for registration back to their original coordinate space
        """
        image_fps = []

        if all(
            [reg_edge.get("registered") for reg_edge in self.reg_graph_edges]
        ):
            # prepare workflow
            merge_modalities = []
            if len(self.merge_modalities.keys()) > 0:
                for k, v in self.merge_modalities.items():
                    merge_modalities.extend(v)

            reg_path_keys = list(self.reg_paths.keys())
            nonreg_keys = self._find_nonreg_modalities()

            if remove_merged:
                for merge_mod in merge_modalities:
                    try:
                        m_idx = reg_path_keys.index(merge_mod)
                        reg_path_keys.pop(m_idx)
                    except ValueError:
                        pass
                    try:
                        m_idx = nonreg_keys.index(merge_mod)
                        nonreg_keys.pop(m_idx)
                    except ValueError:
                        pass

            for modality in reg_path_keys:
                (
                    im_data,
                    transformations,
                    output_path,
                ) = self._prepare_reg_image_transform(
                    modality,
                    attachment=False,
                    to_original_size=to_original_size,
                )

                im_fp = self._transform_write_image(
                    im_data,
                    transformations,
                    output_path,
                    file_writer=file_writer,
                )
                image_fps.append(im_fp)

            for (
                modality,
                attachment_modality,
            ) in self.attachment_images.items():
                if modality in merge_modalities and remove_merged:
                    continue
                if attachment_modality in self._find_nonreg_modalities():
                    (
                        im_data,
                        transformations,
                        output_path,
                    ) = self._prepare_nonreg_image_transform(
                        modality,
                        attachment=True,
                        attachment_modality=attachment_modality,
                        to_original_size=to_original_size,
                    )
                else:
                    (
                        im_data,
                        transformations,
                        output_path,
                    ) = self._prepare_reg_image_transform(
                        modality,
                        attachment=True,
                        attachment_modality=attachment_modality,
                        to_original_size=to_original_size,
                    )

                im_fp = self._transform_write_image(
                    im_data,
                    transformations,
                    output_path,
                    file_writer=file_writer,
                )
                image_fps.append(im_fp)

            if len(self.merge_modalities.items()) > 0:
                im_fp = self._transform_write_merge_images(
                    to_original_size=to_original_size
                )
                image_fps.append(im_fp)

        if transform_non_reg:
            # preprocess and save unregistered nodes
            for modality in nonreg_keys:
                if modality in merge_modalities and remove_merged:
                    continue

                (
                    im_data,
                    transformations,
                    output_path,
                ) = self._prepare_nonreg_image_transform(
                    modality,
                    to_original_size=to_original_size,
                )
                im_fp = self._transform_write_image(
                    im_data,
                    transformations,
                    output_path,
                    file_writer=file_writer,
                )
                image_fps.append(im_fp)

        return image_fps

    def transform_shapes(self):
        """
        Transform all attached shapes and write out shape data to geoJSON.
        """
        transformed_shapes_fps = []

        # do inversion before any processing to ensure it is not computed twice
        for set_name, set_data in self.shape_sets.items():
            attachment_modality = set_data["attachment_modality"]
            invert_nonrigid_transforms(
                self.transformations[attachment_modality][
                    "full-transform-seq"
                ].reg_transforms_itk_order
            )

        for set_name, set_data in self.shape_sets.items():
            attachment_modality = set_data["attachment_modality"]
            im_data = self.modalities[attachment_modality]

            final_modality = self.reg_paths[attachment_modality][-1]

            print(
                "transforming shape set {} associated with {} to {}".format(
                    set_name, attachment_modality, final_modality
                )
            )

            rs = RegShapes(
                set_data["shape_files"], source_res=set_data["image_res"]
            )

            transformations = copy(
                self.transformations[attachment_modality]["full-transform-seq"]
            )

            if im_data["output_res"]:
                transformations.set_output_spacing(im_data["output_res"])

            else:
                output_spacing_target = self.modalities[final_modality][
                    "image_res"
                ]
                transformations.set_output_spacing(
                    (output_spacing_target, output_spacing_target)
                )

            rs.transform_shapes(transformations)

            output_path = (
                self.output_dir
                / "{}-{}-{}_to_{}-transformed_shapes.geojson".format(
                    self.project_name,
                    set_name,
                    attachment_modality,
                    final_modality,
                )
            )

            rs.save_shape_data(output_path, transformed=True)
            transformed_shapes_fps.append(output_path)

        return transformed_shapes_fps

    def _transforms_to_txt(
        self, transformations: Dict[str, RegTransformSeq]
    ) -> Dict[str, List[str]]:
        tform_txt = dict()

        for k, v in transformations.items():
            if k == "full-transform-seq":
                continue
            if "initial" in k:
                if v:
                    for idx, rt in enumerate(v.reg_transforms):
                        tform_txt.update(
                            {f"{k}-{str(idx).zfill(2)}": rt.elastix_transform}
                        )
            else:
                tform_txt.update(
                    {f"{k}": [rt.elastix_transform for rt in v.reg_transforms]}
                )

        return tform_txt

    def save_transformations(self):
        """
        Save all transformations for a given modality as JSON
        """
        if all(
            [reg_edge.get("registered") for reg_edge in self.reg_graph_edges]
        ):
            for key in self.reg_paths.keys():

                final_modality = self.reg_paths[key][-1]

                output_path = (
                    self.output_dir
                    / "{}-{}_to_{}_transformations.json".format(
                        self.project_name,
                        key,
                        final_modality,
                    )
                )
                tform_txt = self._transforms_to_txt(self.transformations[key])

                with open(output_path, 'w') as fp:
                    json.dump(tform_txt, fp, indent=1)

            for (
                modality,
                attachment_modality,
            ) in self.attachment_images.items():
                if attachment_modality not in self._find_nonreg_modalities():

                    final_modality = self.reg_paths[attachment_modality][-1]

                    output_path = (
                        self.output_dir
                        / "{}-{}_to_{}_transformations.json".format(
                            self.project_name,
                            modality,
                            final_modality,
                        )
                    )

                    tform_txt = self._transforms_to_txt(
                        self.transformations[key]
                    )

                    with open(output_path, 'w') as fp:
                        json.dump(tform_txt, fp, indent=4)
        else:
            warn(
                "registration has not been executed for the graph "
                "no transformations to save"
            )

    def add_data_from_config(self, config_filepath):

        reg_config = parse_check_reg_config(config_filepath)

        if reg_config.get("modalities"):
            for key, val in reg_config["modalities"].items():

                image_filepath = (
                    val.get("image_filepath")
                    if val.get("image_filepath") is not None
                    else val.get("image_filepath")
                )

                preprocessing = (
                    "None"
                    if val.get("preprocessing") is None
                    else val.get("preprocessing")
                )

                self.add_modality(
                    key,
                    image_filepath,
                    image_res=val.get("image_res"),
                    channel_names=val.get("channel_names"),
                    channel_colors=val.get("channel_colors"),
                    preprocessing=preprocessing,
                    mask=val.get("mask"),
                    output_res=val.get("output_res"),
                )
        else:
            print("warning: config file did not contain any image modalities")

        if reg_config.get("reg_paths"):

            for key, val in reg_config["reg_paths"].items():
                self.add_reg_path(
                    val.get("src_modality_name"),
                    val.get("tgt_modality_name"),
                    val.get("thru_modality"),
                    reg_params=val.get("reg_params"),
                    override_prepro=val.get("override_prepro"),
                )
        else:
            print(
                "warning: config file did not contain any registration paths"
            )
        if reg_config.get("attachment_images"):

            for key, val in reg_config["attachment_images"].items():
                self.add_attachment_images(
                    val.get("attachment_modality"),
                    key,
                    val.get("image_filepath"),
                    val.get("image_res"),
                    channel_names=val.get("channel_names"),
                    channel_colors=val.get("channel_colors"),
                )

        if reg_config.get("attachment_shapes"):

            for key, val in reg_config["attachment_shapes"].items():
                self.add_attachment_shapes(
                    val.get("attachment_modality"), key, val.get("shape_files")
                )

        if reg_config.get("reg_graph_edges"):
            self._reg_graph_edges = reg_config["reg_graph_edges"]
            if all([re.get("registered") for re in self.reg_graph_edges]):
                self.transformations = self.reg_graph_edges

        if reg_config.get("merge_modalities"):
            for mn, mm in reg_config["merge_modalities"].items():
                self.add_merge_modalities(mn, mm)

    def reset_registered_modality(self, modalities):
        edge_keys = [
            r.get("modalities").get("source") for r in self.reg_graph_edges
        ]
        if isinstance(modalities, str):
            modalities = [modalities]

        for modality in modalities:
            modality_idx = edge_keys.index(modality)
            self.reg_graph_edges[modality_idx]["registered"] = False


def main():
    import argparse

    def config_to_WsiReg2D(config_filepath):
        reg_config = parse_check_reg_config(config_filepath)

        reg_graph = WsiReg2D(
            reg_config.get("project_name"),
            reg_config.get("output_dir"),
            reg_config.get("cache_images"),
        )
        return reg_graph

    parser = argparse.ArgumentParser(
        description='Load Whole Slide Image 2D Registration Graph from configuration file'
    )

    parser.add_argument(
        "config_filepath",
        metavar="C",
        type=str,
        nargs=1,
        help="full filepath for .yaml configuration file",
    )
    parser.add_argument(
        "--fw",
        type=str,
        nargs=1,
        help="how to write output registered images: ome.tiff, ome.zarr (default: ome.tiff)",
    )

    parser.add_argument('--write_im', dest='write_im', action='store_true')
    parser.add_argument('--no_write_im', dest='write_im', action='store_false')
    parser.set_defaults(write_im=True)

    args = parser.parse_args()
    config_filepath = args.config_filepath[0]
    if args.fw is None:
        file_writer = "ome.tiff"
    else:
        file_writer = args.fw[0]

    reg_graph = config_to_WsiReg2D(config_filepath)
    reg_graph.add_data_from_config(config_filepath)

    reg_graph.register_images()
    reg_graph.save_transformations()

    if args.write_im:
        reg_graph.transform_images(file_writer=file_writer)

    if reg_graph.shape_sets:
        reg_graph.transform_shapes()


if __name__ == "__main__":
    import sys

    sys.exit(main())



from .reg_image import RegImage  # noqa: F401
from .tifffile_reg_image import TiffFileRegImage  # noqa: F401
from .czi_reg_image import CziRegImage  # noqa: F401
from .ometiff_reg_image import OmeTiffRegImage  # noqa: F401
from .sitk_reg_image import SitkRegImage  # noqa: F401
from .aics_reg_image import AICSRegImage  # noqa: F401
from .np_reg_image import NumpyRegImage  # noqa: F401
from .merge_reg_image import MergeRegImage  # noqa: F401



import os
import random
import string
from pathlib import Path

import numpy as np
import pytest
from ome_types import from_xml
from tifffile import TiffFile

from wsireg.parameter_maps.preprocessing import ImagePreproParams
from wsireg.reg_images.loader import reg_image_loader
from wsireg.wsireg2d import WsiReg2D

HERE = os.path.dirname(__file__)
GEOJSON_FP = os.path.join(HERE, "fixtures/polygons.geojson")


def gen_project_name_str():
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(10))


@pytest.fixture(scope="session")
def data_out_dir(tmpdir_factory):
    out_dir = tmpdir_factory.mktemp("output")
    return out_dir


@pytest.fixture(scope="session")
def data_im_fp(tmpdir_factory):
    out_im = tmpdir_factory.mktemp("image").join("image_fp.tiff")
    return out_im


def test_WsiReg2D_instantiation(data_out_dir):
    pstr = gen_project_name_str()
    wsi_reg = WsiReg2D(pstr, str(data_out_dir))
    assert wsi_reg.project_name == pstr
    assert wsi_reg.output_dir == Path(str(data_out_dir))


def test_wsireg2d_add_modality_w_fp(data_out_dir, data_im_fp):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(data_im_fp)
    wsi_reg.add_modality(
        "test_mod",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )
    assert wsi_reg.modalities.get("test_mod").get("image_filepath") == img_fp1
    assert wsi_reg.modalities.get("test_mod").get("image_res") == 0.65
    assert wsi_reg.modalities.get("test_mod").get("channel_names") == ["test"]
    assert wsi_reg.modalities.get("test_mod").get("channel_colors") == ["red"]
    assert (
        wsi_reg.modalities.get("test_mod").get("preprocessing")
        == ImagePreproParams()
    )
    assert wsi_reg.modalities.get("test_mod").get("mask") == None
    assert wsi_reg.n_modalities == 1


@pytest.mark.usefixtures("im_mch_np")
def test_wsireg2d_add_modality_w_np(data_out_dir, im_mch_np):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    wsi_reg.add_modality(
        "test_mod",
        im_mch_np,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )
    assert np.array_equal(
        wsi_reg.modalities.get("test_mod").get("image_filepath"), im_mch_np
    )
    assert wsi_reg.modalities.get("test_mod").get("image_res") == 0.65
    assert wsi_reg.modalities.get("test_mod").get("channel_names") == ["test"]
    assert wsi_reg.modalities.get("test_mod").get("channel_colors") == ["red"]
    assert (
        wsi_reg.modalities.get("test_mod").get("preprocessing")
        == ImagePreproParams()
    )
    assert wsi_reg.modalities.get("test_mod").get("mask") == None
    assert wsi_reg.n_modalities == 1


def test_wsireg2d_add_modality_check_names(data_out_dir, data_im_fp):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(data_im_fp)
    modality_name1 = "test_mod1"
    modality_name2 = "test_mod2"
    wsi_reg.add_modality(
        modality_name1,
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )
    wsi_reg.add_modality(
        modality_name2,
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )
    assert modality_name1 in wsi_reg.modality_names
    assert modality_name2 in wsi_reg.modality_names
    assert wsi_reg.modality_names == ["test_mod1", "test_mod2"]
    assert wsi_reg.n_modalities == 2


def test_wsireg2d_add_reg_path_single(data_out_dir, data_im_fp):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(data_im_fp)
    modality_name1 = "test_mod1"
    modality_name2 = "test_mod2"
    wsi_reg.add_modality(
        modality_name1,
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )
    wsi_reg.add_modality(
        modality_name2,
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )
    wsi_reg.add_reg_path(modality_name1, modality_name2, reg_params=["rigid"])
    assert wsi_reg.n_registrations == 1
    assert wsi_reg.reg_paths.get(modality_name1) == [modality_name2]
    assert (
        wsi_reg.reg_graph_edges[0].get("modalities").get("target")
        == modality_name2
    )


def test_wsireg2d_add_modality_duplicated_error(data_out_dir, data_im_fp):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(data_im_fp)
    wsi_reg.add_modality(
        "test_mod",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )
    with pytest.raises(ValueError):
        wsi_reg.add_modality(
            "test_mod",
            img_fp1,
            0.65,
            channel_names=["test"],
            channel_colors=["red"],
        )


def test_wsireg2d_add_merge_modality_notfound(data_out_dir, data_im_fp):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(data_im_fp)
    wsi_reg.add_modality(
        "test_mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )
    img_fp1 = str(data_im_fp)
    wsi_reg.add_modality(
        "test_mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )
    with pytest.raises(ValueError):
        wsi_reg.add_merge_modalities("mergetest", ["test_mod1", "test_mod3"])


@pytest.mark.usefixtures("disk_im_gry")
def test_wsireg_run_reg(data_out_dir, disk_im_gry):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(disk_im_gry)

    wsi_reg.add_modality(
        "mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )

    wsi_reg.add_modality(
        "mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )

    wsi_reg.add_reg_path(
        "mod1", "mod2", reg_params=["rigid_test", "affine_test"]
    )
    wsi_reg.register_images()
    wsi_reg.save_transformations()
    im_fps = wsi_reg.transform_images(transform_non_reg=True)

    assert Path(im_fps[0]).exists() is True
    assert Path(im_fps[1]).exists() is True


@pytest.mark.usefixtures("disk_im_gry")
def test_wsireg_run_reg_with_crop(data_out_dir, disk_im_gry):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(disk_im_gry)

    wsi_reg.add_modality(
        "mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )

    wsi_reg.add_modality(
        "mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={
            "mask_bbox": [512, 512, 512, 512],
            "crop_to_mask_bbox": True,
        },
    )

    wsi_reg.add_reg_path(
        "mod1", "mod2", reg_params=["rigid_test", "affine_test"]
    )
    wsi_reg.register_images()
    # not cropped
    im_fps = wsi_reg.transform_images(
        transform_non_reg=True, to_original_size=True
    )

    registered_image_nocrop = reg_image_loader(im_fps[0], 1)
    unregistered_image_nocrop = reg_image_loader(im_fps[1], 1)

    # crop image
    im_fps = wsi_reg.transform_images(
        transform_non_reg=True, to_original_size=False
    )
    wsi_reg.save_transformations()

    #
    registered_image_crop = reg_image_loader(im_fps[0], 1)
    unregistered_image_crop = reg_image_loader(im_fps[1], 1)
    #
    assert registered_image_nocrop.im_dims[1:] == (2048, 2048)
    assert unregistered_image_nocrop.im_dims[1:] == (2048, 2048)
    assert registered_image_crop.im_dims[1:] == (512, 512)
    assert unregistered_image_crop.im_dims[1:] == (512, 512)


@pytest.mark.usefixtures("disk_im_gry")
def test_wsireg_run_reg_with_flip_crop(data_out_dir, disk_im_gry):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(disk_im_gry)

    wsi_reg.add_modality(
        "mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )

    wsi_reg.add_modality(
        "mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={
            "mask_bbox": [512, 512, 512, 512],
            "flip": "h",
            "crop_to_mask_bbox": True,
        },
    )

    wsi_reg.add_reg_path("mod1", "mod2", reg_params=["rigid_test"])
    wsi_reg.register_images()
    # not cropped
    im_fps = wsi_reg.transform_images(
        transform_non_reg=True, to_original_size=True
    )
    registered_image_nocrop = reg_image_loader(im_fps[0], 1)
    unregistered_image_nocrop = reg_image_loader(im_fps[1], 1)

    # crop image
    im_fps = wsi_reg.transform_images(
        transform_non_reg=True, to_original_size=False
    )
    wsi_reg.save_transformations()

    registered_image_crop = reg_image_loader(im_fps[0], 1)
    unregistered_image_crop = reg_image_loader(im_fps[1], 1)

    assert registered_image_nocrop.im_dims[1:] == (2048, 2048)
    assert unregistered_image_nocrop.im_dims[1:] == (2048, 2048)
    assert registered_image_crop.im_dims[1:] == (512, 512)
    assert unregistered_image_crop.im_dims[1:] == (512, 512)


@pytest.mark.usefixtures("disk_im_gry")
def test_wsireg_run_reg_with_crop_merge(data_out_dir, disk_im_gry):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(disk_im_gry)

    wsi_reg.add_modality(
        "mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )

    wsi_reg.add_modality(
        "mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )

    wsi_reg.add_modality(
        "mod3",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={
            "mask_bbox": [512, 512, 512, 512],
            "crop_to_mask_bbox": True,
        },
    )

    wsi_reg.add_reg_path("mod1", "mod3", reg_params=["rigid_test"])
    wsi_reg.add_reg_path("mod2", "mod3", reg_params=["rigid_test"])
    wsi_reg.register_images()
    wsi_reg.add_merge_modalities("merge", ["mod1", "mod2", "mod3"])
    # not cropped
    im_fps = wsi_reg.transform_images(
        transform_non_reg=True, to_original_size=True
    )
    registered_image_nocrop = reg_image_loader(im_fps[0], 1)

    # crop image
    im_fps = wsi_reg.transform_images(
        transform_non_reg=True, to_original_size=False
    )
    wsi_reg.save_transformations()
    registered_image_crop = reg_image_loader(im_fps[0], 1)
    assert registered_image_nocrop.im_dims[1:] == (2048, 2048)
    assert registered_image_crop.im_dims[1:] == (512, 512)


def test_wsireg_run_reg_wmerge(data_out_dir, disk_im_gry):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(disk_im_gry)

    wsi_reg.add_modality(
        "mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )

    wsi_reg.add_modality(
        "mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )

    wsi_reg.add_reg_path(
        "mod1", "mod2", reg_params=["rigid_test", "affine_test"]
    )
    wsi_reg.add_merge_modalities("test_merge", ["mod1", "mod2"])
    wsi_reg.register_images()
    wsi_reg.save_transformations()
    im_fps = wsi_reg.transform_images(transform_non_reg=True)
    merged_im = reg_image_loader(im_fps[0], 0.65)
    assert Path(im_fps[0]).exists() is True
    assert merged_im.im_dims == (2, 2048, 2048)


def test_wsireg_run_reg_wmerge_and_indiv(data_out_dir, disk_im_gry):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(disk_im_gry)

    wsi_reg.add_modality(
        "mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )

    wsi_reg.add_modality(
        "mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )

    wsi_reg.add_modality(
        "mod3",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )

    wsi_reg.add_reg_path(
        "mod2", "mod1", reg_params=["rigid_test", "affine_test"]
    )
    wsi_reg.add_reg_path(
        "mod3", "mod1", reg_params=["rigid_test", "affine_test"]
    )
    wsi_reg.add_merge_modalities("test_merge", ["mod1", "mod2"])
    wsi_reg.register_images()
    im_fps = wsi_reg.transform_images(
        remove_merged=False, transform_non_reg=True
    )

    merged_im = reg_image_loader(im_fps[-2], 0.65)
    wsi_reg.save_transformations()

    assert len(im_fps) == 4
    assert Path(im_fps[0]).exists() is True
    assert Path(im_fps[1]).exists() is True
    assert Path(im_fps[2]).exists() is True
    assert Path(im_fps[3]).exists() is True
    assert merged_im.im_dims == (2, 2048, 2048)


def test_wsireg_run_reg_wattachment(data_out_dir, disk_im_gry):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    im1 = np.random.randint(0, 255, (2048, 2048), dtype=np.uint16)
    im2 = np.random.randint(0, 255, (2048, 2048), dtype=np.uint16)

    wsi_reg.add_modality(
        "mod1",
        im1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )

    wsi_reg.add_modality(
        "mod2",
        im2,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )
    wsi_reg.add_attachment_images("mod2", "attached", im2, image_res=0.65)
    wsi_reg.add_attachment_images("mod1", "attached2", im1, image_res=0.65)

    wsi_reg.add_reg_path(
        "mod2", "mod1", reg_params=["rigid_test", "affine_test"]
    )

    wsi_reg.register_images()
    im_fps = wsi_reg.transform_images(transform_non_reg=False)

    wsi_reg.save_transformations()

    regim = reg_image_loader(im_fps[0], 0.65)
    attachim = reg_image_loader(im_fps[1], 0.65)
    attachim2 = reg_image_loader(im_fps[2], 0.65)

    assert np.array_equal(regim.image.compute(), attachim.image.compute())
    assert np.array_equal(im1, attachim2.image.compute())


def test_wsireg_run_reg_wattachment_ds2(data_out_dir, disk_im_gry):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    im1 = np.random.randint(0, 255, (2048, 2048), dtype=np.uint16)
    im2 = np.random.randint(0, 255, (2048, 2048), dtype=np.uint16)

    wsi_reg.add_modality(
        "mod1",
        im1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"downsampling": 2},
    )

    wsi_reg.add_modality(
        "mod2",
        im2,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"downsampling": 2},
    )
    wsi_reg.add_attachment_images("mod2", "attached", im2, image_res=0.65)
    wsi_reg.add_attachment_images("mod1", "attached2", im1, image_res=0.65)

    wsi_reg.add_reg_path(
        "mod2", "mod1", reg_params=["rigid_test", "affine_test"]
    )

    wsi_reg.register_images()
    im_fps = wsi_reg.transform_images(transform_non_reg=False)

    wsi_reg.save_transformations()

    regim = reg_image_loader(im_fps[0], 0.65)
    attachim = reg_image_loader(im_fps[1], 0.65)
    attachim2 = reg_image_loader(im_fps[2], 0.65)

    assert np.array_equal(regim.image.compute(), attachim.image.compute())
    assert np.array_equal(im1, attachim2.image.compute())


@pytest.mark.usefixtures("disk_im_gry")
def test_wsireg_run_reg_shapes(data_out_dir, disk_im_gry):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(disk_im_gry)

    wsi_reg.add_modality(
        "mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )

    wsi_reg.add_modality(
        "mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )

    wsi_reg.add_reg_path(
        "mod1", "mod2", reg_params=["rigid_test", "affine_test"]
    )
    wsi_reg.add_attachment_shapes("mod1", "shapeset", GEOJSON_FP)
    wsi_reg.register_images()
    wsi_reg.transform_shapes()
    wsi_reg.save_transformations()
    im_fps = wsi_reg.transform_images(transform_non_reg=False)
    gj_files = sorted(Path(im_fps[0]).parent.glob("*.geojson"))

    assert Path(im_fps[0]).exists() is True
    assert len(gj_files) > 0


@pytest.mark.usefixtures("disk_im_gry")
def test_wsireg_run_reg_changeres(data_out_dir, disk_im_gry):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(disk_im_gry)

    wsi_reg.add_modality(
        "mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        output_res=0.325,
    )

    wsi_reg.add_modality(
        "mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )

    wsi_reg.add_reg_path(
        "mod1", "mod2", reg_params=["rigid_test", "affine_test"]
    )
    wsi_reg.register_images()

    im_fps = wsi_reg.transform_images(transform_non_reg=False)
    regim = reg_image_loader(im_fps[0], 0.325)

    assert regim.im_dims[1:] == (4096, 4096)


@pytest.mark.usefixtures("disk_im_gry")
def test_wsireg_run_reg_downsampling_m1(data_out_dir, disk_im_gry):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(disk_im_gry)

    wsi_reg.add_modality(
        "mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"downsampling": 2},
    )

    wsi_reg.add_modality(
        "mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )

    wsi_reg.add_reg_path(
        "mod1", "mod2", reg_params=["rigid_test", "affine_test"]
    )
    wsi_reg.register_images()

    im_fps = wsi_reg.transform_images(transform_non_reg=False)
    regim = reg_image_loader(im_fps[0], 0.65)

    assert regim.im_dims[1:] == (2048, 2048)


@pytest.mark.usefixtures("disk_im_gry")
def test_wsireg_run_reg_downsampling_m1_prepro(data_out_dir, disk_im_gry):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(disk_im_gry)

    wsi_reg.add_modality(
        "mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"downsampling": 2, "rot_cc": 90},
    )

    wsi_reg.add_modality(
        "mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )

    wsi_reg.add_reg_path(
        "mod1", "mod2", reg_params=["rigid_test", "affine_test"]
    )
    wsi_reg.register_images()

    im_fps = wsi_reg.transform_images(transform_non_reg=False)
    regim = reg_image_loader(im_fps[0], 0.65)

    assert regim.im_dims[1:] == (2048, 2048)


@pytest.mark.usefixtures("disk_im_gry")
def test_wsireg_run_reg_downsampling_m1m2(data_out_dir, disk_im_gry):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(disk_im_gry)

    wsi_reg.add_modality(
        "mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"downsampling": 2},
    )

    wsi_reg.add_modality(
        "mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"downsampling": 2},
    )

    wsi_reg.add_reg_path("mod1", "mod2", reg_params=["rigid_test"])
    wsi_reg.register_images()

    im_fps = wsi_reg.transform_images(transform_non_reg=False)
    regim = reg_image_loader(im_fps[0], 0.65)

    assert regim.im_dims[1:] == (2048, 2048)


@pytest.mark.usefixtures("disk_im_gry")
def test_wsireg_run_reg_downsampling_m1m2_changeores(
    data_out_dir, disk_im_gry
):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(disk_im_gry)

    wsi_reg.add_modality(
        "mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"downsampling": 2},
        output_res=(1.3, 1.3),
    )

    wsi_reg.add_modality(
        "mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"downsampling": 2},
    )

    wsi_reg.add_reg_path("mod1", "mod2", reg_params=["rigid_test"])
    wsi_reg.register_images()

    im_fps = wsi_reg.transform_images(transform_non_reg=False)
    regim = reg_image_loader(im_fps[0], 0.65)

    assert regim.im_dims[1:] == (1024, 1024)


@pytest.mark.usefixtures("disk_im_gry")
def test_wsireg_run_reg_downsampling_m2_prepro(data_out_dir, disk_im_gry):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(disk_im_gry)

    wsi_reg.add_modality(
        "mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"downsampling": 2},
    )

    wsi_reg.add_modality(
        "mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"rot_cc": 90, "downsampling": 2},
    )

    wsi_reg.add_reg_path("mod1", "mod2", reg_params=["rigid_test"])
    wsi_reg.register_images()

    im_fps = wsi_reg.transform_images(transform_non_reg=True)
    regim = reg_image_loader(im_fps[1], 0.65)

    assert regim.im_dims[1:] == (2048, 2048)


@pytest.mark.usefixtures("disk_im_gry")
def test_wsireg_run_reg_downsampling_m1m2_merge(data_out_dir, disk_im_gry):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(disk_im_gry)

    wsi_reg.add_modality(
        "mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"downsampling": 2, "rot_cc": 90},
    )

    wsi_reg.add_modality(
        "mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"rot_cc": 90, "downsampling": 2},
    )

    wsi_reg.add_reg_path("mod1", "mod2", reg_params=["rigid_test"])
    wsi_reg.add_merge_modalities("mod12-merge", ["mod1", "mod2"])
    wsi_reg.register_images()

    im_fps = wsi_reg.transform_images(
        transform_non_reg=True, remove_merged=True
    )
    regim = reg_image_loader(im_fps[0], 0.65)

    assert regim.im_dims == (2, 2048, 2048)


@pytest.mark.usefixtures("disk_im_gry")
def test_wsireg_run_reg_downsampling_m1m2_merge_no_prepro(
    data_out_dir, disk_im_gry
):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(disk_im_gry)

    wsi_reg.add_modality(
        "mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"downsampling": 2},
    )

    wsi_reg.add_modality(
        "mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"downsampling": 2},
    )

    wsi_reg.add_reg_path("mod1", "mod2", reg_params=["rigid_test"])
    wsi_reg.add_merge_modalities("mod12-merge", ["mod1", "mod2"])
    wsi_reg.register_images()

    im_fps = wsi_reg.transform_images(
        transform_non_reg=False, remove_merged=True
    )
    regim = reg_image_loader(im_fps[0], 0.65)

    assert regim.im_dims == (2, 2048, 2048)


@pytest.mark.usefixtures("disk_im_gry")
def test_wsireg_run_reg_downsampling_m1m2_merge_ds_attach(
    data_out_dir, disk_im_gry
):
    wsi_reg = WsiReg2D(gen_project_name_str(), str(data_out_dir))
    img_fp1 = str(disk_im_gry)

    wsi_reg.add_modality(
        "mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"downsampling": 2},
    )

    wsi_reg.add_modality(
        "mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"downsampling": 2},
    )

    wsi_reg.add_attachment_images(
        "mod2",
        "mod3",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )
    wsi_reg.add_reg_path("mod1", "mod2", reg_params=["rigid_test"])

    wsi_reg.add_merge_modalities("mod12-merge", ["mod1", "mod2", "mod3"])
    wsi_reg.register_images()

    im_fps = wsi_reg.transform_images(
        transform_non_reg=False, remove_merged=True
    )
    regim = reg_image_loader(im_fps[0], 0.65)
    ome_data = from_xml(TiffFile(im_fps[0]).ome_metadata)

    assert regim.im_dims == (3, 2048, 2048)
    assert ome_data.images[0].pixels.physical_size_x == 0.65
    assert ome_data.images[0].pixels.physical_size_y == 0.65
    assert ome_data.images[0].pixels.size_x == 2048
    assert ome_data.images[0].pixels.size_y == 2048
    assert ome_data.images[0].pixels.size_c == 3


@pytest.mark.usefixtures("disk_im_gry")
def test_wsireg_run_reg_downsampling_from_cache(data_out_dir, disk_im_gry):
    pstr = gen_project_name_str()
    wsi_reg = WsiReg2D(pstr, str(data_out_dir))
    img_fp1 = str(disk_im_gry)

    wsi_reg.add_modality(
        "mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"downsampling": 2},
    )

    wsi_reg.add_modality(
        "mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"downsampling": 2},
    )

    wsi_reg.add_attachment_images(
        "mod2",
        "mod3",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )
    wsi_reg.add_reg_path("mod1", "mod2", reg_params=["rigid_test"])
    wsi_reg.register_images()

    im_fps_nocache = wsi_reg.transform_images(
        transform_non_reg=True, remove_merged=True
    )

    regim_nocache = reg_image_loader(im_fps_nocache[0], 0.65)
    regim_nocache_attach = reg_image_loader(im_fps_nocache[1], 0.65)
    regim_nocache_br = reg_image_loader(im_fps_nocache[2], 0.65)

    wsi_reg2 = WsiReg2D(pstr, str(data_out_dir))
    img_fp1 = str(disk_im_gry)

    wsi_reg2.add_modality(
        "mod1",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"downsampling": 2},
    )

    wsi_reg2.add_modality(
        "mod2",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
        preprocessing={"downsampling": 2},
    )

    wsi_reg2.add_attachment_images(
        "mod2",
        "mod3",
        img_fp1,
        0.65,
        channel_names=["test"],
        channel_colors=["red"],
    )
    wsi_reg2.add_reg_path("mod1", "mod2", reg_params=["rigid_test"])
    wsi_reg2.register_images()

    im_fps_cache = wsi_reg.transform_images(
        transform_non_reg=True, remove_merged=True
    )

    regim_cache = reg_image_loader(im_fps_cache[0], 0.65)
    regim_cache_attach = reg_image_loader(im_fps_cache[1], 0.65)
    regim_cache_br = reg_image_loader(im_fps_cache[2], 0.65)

    regim_nocache_attach.im_dims
    regim_cache_attach.im_dims

    assert regim_cache.im_dims == regim_nocache.im_dims
    assert regim_cache_br.im_dims == regim_nocache_br.im_dims
    assert regim_cache_attach.im_dims == regim_nocache_attach.im_dims



import os
from pathlib import Path
from typing import Union

import cv2
import numpy as np
import pytest
import SimpleITK as sitk

from wsireg.reg_images.loader import reg_image_loader
from wsireg.reg_shapes import RegShapes
from wsireg.utils.config_utils import parse_check_reg_config
from wsireg.wsireg2d import WsiReg2D

HERE = os.path.dirname(__file__)
FIXTURES_DIR = os.path.join(HERE, "fixtures")
PRIVATE_DIR = os.path.join(HERE, "private_data")

config1_fp = str(Path(FIXTURES_DIR) / "test-config1.yaml")
config2_fp = str(Path(FIXTURES_DIR) / "test-config2.yaml")
config3_fp = str(Path(FIXTURES_DIR) / "test-config3.yaml")
config4_fp = str(Path(FIXTURES_DIR) / "test-config4.yaml")
config5_fp = str(Path(FIXTURES_DIR) / "test-config5.yaml")

SKIP_PRIVATE = False
REASON = "private data"

if not os.path.exists(PRIVATE_DIR):
    SKIP_PRIVATE = True


def geojson_to_binary_image(geojson_fp: Union[str, Path]) -> sitk.Image:
    rs = RegShapes(geojson_fp)
    shape_cv2 = [s["array"].astype(np.int32) for s in rs.shape_data]

    x_max = np.max([np.max(s["array"][:, 0]) for s in rs.shape_data]).astype(
        int
    )
    y_max = np.max([np.max(s["array"][:, 1]) for s in rs.shape_data]).astype(
        int
    )

    binary_image = np.zeros((y_max + 200, x_max + 200), dtype=np.uint8)
    return sitk.GetImageFromArray(cv2.fillPoly(binary_image, shape_cv2, 255))


def compute_dice(b1: sitk.Image, b2: sitk.Image):

    b2 = sitk.Resample(
        b2,
        b1.GetSize(),
        sitk.Transform(),
        sitk.sitkNearestNeighbor,
        b1.GetOrigin(),
        b1.GetSpacing(),
        b1.GetDirection(),
        0,
        b2.GetPixelID(),
    )
    labstats = sitk.LabelOverlapMeasuresImageFilter()

    labstats.Execute(b1, b2)

    return labstats.GetDiceCoefficient()


def config_to_WsiReg2D(config_filepath, output_dir):
    reg_config = parse_check_reg_config(config_filepath)

    reg_graph = WsiReg2D(
        reg_config.get("project_name"),
        output_dir,
        reg_config.get("cache_images"),
    )
    return reg_graph


@pytest.fixture(scope="session")
def data_out_dir(tmpdir_factory):
    out_dir = tmpdir_factory.mktemp("output")
    return out_dir


@pytest.mark.parametrize("config_fp", [(config1_fp), (config2_fp)])
def test_wsireg_configs(config_fp, data_out_dir):
    wsi_reg = config_to_WsiReg2D(config_fp, data_out_dir)
    wsi_reg.add_data_from_config(config_fp)
    wsi_reg.register_images()
    wsi_reg.save_transformations()
    assert wsi_reg.output_dir == Path(str(data_out_dir))


@pytest.mark.parametrize("config_fp", [(config1_fp), (config2_fp)])
def test_wsireg_configs_fromcache(config_fp, data_out_dir):
    wsi_reg1 = config_to_WsiReg2D(config_fp, data_out_dir)
    wsi_reg1.add_data_from_config(config_fp)
    wsi_reg1.register_images()

    wsi_reg2 = config_to_WsiReg2D(config_fp, data_out_dir)
    wsi_reg2.add_data_from_config(config_fp)
    wsi_reg2.register_images()

    assert wsi_reg1.output_dir == Path(str(data_out_dir))


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
@pytest.mark.parametrize(
    "config_fp",
    [
        (config3_fp),
    ],
)
def test_wsireg_config_full_exp_DICE(config_fp, data_out_dir):
    wsi_reg1 = config_to_WsiReg2D(config_fp, data_out_dir)
    wsi_reg1.add_data_from_config(config_fp)
    wsi_reg1.register_images()
    shape_fps = wsi_reg1.transform_shapes()
    gt = geojson_to_binary_image(
        "private_data/unreg_rois/VAN0006-LK-2-85-AF_preIMS_unregistered.geojson"
    )

    dice_vals = []
    for shape in shape_fps:
        test_mask = geojson_to_binary_image(shape)
        dice_vals.append(compute_dice(gt, test_mask))

    assert all(np.asarray(dice_vals) > 0.85)


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
@pytest.mark.parametrize(
    "config_fp",
    [
        (config4_fp),
    ],
)
def test_wsireg_config_full_exp_DICE_ds(config_fp, data_out_dir):
    wsi_reg1 = config_to_WsiReg2D(config_fp, data_out_dir)
    wsi_reg1.add_data_from_config(config_fp)
    wsi_reg1.register_images()
    shape_fps = wsi_reg1.transform_shapes()
    gt = geojson_to_binary_image(
        "private_data/unreg_rois/VAN0006-LK-2-85-AF_preIMS_unregistered.geojson"
    )

    dice_vals = []
    for shape in shape_fps:
        test_mask = geojson_to_binary_image(shape)
        dice_vals.append(compute_dice(gt, test_mask))

    assert all(np.asarray(dice_vals) > 0.8)


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
@pytest.mark.parametrize(
    "config_fp",
    [
        (config5_fp),
    ],
)
def test_wsireg_config_full_merge_rgb_mc(config_fp, data_out_dir):
    wsi_reg1 = config_to_WsiReg2D(config_fp, data_out_dir)
    wsi_reg1.add_data_from_config(config_fp)
    wsi_reg1.register_images()
    im_fps = wsi_reg1.transform_images()
    ri = reg_image_loader(im_fps[0], 1)

    assert ri.im_dtype == np.uint16
    assert ri.im_dims == (9, 3993, 3397)



import json
from pathlib import Path
from typing import Tuple, Union

import cv2
import numpy as np

from wsireg.reg_transform_seq import RegTransformSeq
from wsireg.utils.shape_utils import (
    get_int_dtype,
    insert_transformed_pts_gj,
    invert_nonrigid_transforms,
    scale_shape_coordinates,
    shape_reader,
    transform_shapes,
)


class RegShapes:
    """
    Class that holds and manages shape data and it's transformations in the registration graph

    Parameters
    ----------
    shape_data: list
        list of np.ndarrays of shape data
        str - file path to GeoJSON file containing shape data
        np.ndarray - single shape
    itk_point_transforms: list
        list of ITK point transforms, appropriately inverted where non-linear.
        Usually generated on-the-fly when points are transformed from wsireg transformation data
    source_res:float
        isotropic image resolution of the source image in the registration stack,
        i.e., resolution of the image to which shape data is associated
    target_res:float
        isotropic image resolution of the target image in the registration stack
        usually generated on-the-fly when points are transformed from wsireg transformation data
    kwargs
        keyword arguments passed to wsireg.utils.tform_utils.shape_reader (shape_name, shape_type)
    """

    def __init__(
        self,
        shape_data=None,
        itk_point_transforms=None,
        source_res=1,
        target_res=1,
        **kwargs,
    ):
        self.shape_data = []
        self.shape_data_gj = []
        self.transformed_shape_data = []
        self.itk_point_transforms = itk_point_transforms
        self.source_res = source_res
        self.target_res = target_res

        self._n_shapes = None
        self._shape_types = None
        self._shape_names = None

        if shape_data:
            self.add_shapes(shape_data, **kwargs)

    @property
    def n_shapes(self):
        """
        Number of shapes loaded
        """
        return self._n_shapes

    @property
    def shape_types(self):
        """
        List of GeoJSON shape types in shape data
        """
        return self._n_shape_types

    @property
    def shape_names(self):
        """
        list of GeoJSON shape names in shape data
        """
        return self._shape_names

    def add_shapes(self, shape_data, **kwargs):
        """
        Add shapes via shape_reader, will extend current shape list rather than overwrite it

        Parameters
        ----------
        shape_data
            list of np.ndarrays of shape data
            str - file path to GeoJSON file containing shape data
            np.ndarray - single shape
        """
        gj_shapes, np_shapes = shape_reader(shape_data, **kwargs)

        self.update_shapes(np_shapes)
        self.update_shapes_gj(gj_shapes)

    def update_shapes(self, imported_shapes):
        """
        Extend list of shape data with new shape data
        """
        self.shape_data.extend(imported_shapes)

    def update_shapes_gj(self, imported_shapes):
        """
        extend list of shape data in GeoJSON format and re-tabulate shape meta data
        """
        self.shape_data_gj.extend(imported_shapes)

        self._n_shapes = len(self.shape_data_gj)
        self._n_shape_types = [
            sh["geometry"]["type"] for sh in self.shape_data_gj
        ]
        self._shape_names = [
            sh["properties"]["classification"]["name"]
            for sh in self.shape_data_gj
        ]

    def scale_shapes(self, scale_factor):
        """
        scale coordinates of list of shape data by scale_factor

        Parameters
        ----------
        scale_factor: float
            isotropic scaling factor for the coordinates
        """
        self.shape_data = [
            scale_shape_coordinates(shape, scale_factor)
            for shape in self.shape_data
        ]

    def transform_shapes(
        self,
        transformations: Union[str, Path, dict, RegTransformSeq],
        px_idx=True,
        output_idx=True,
    ):
        """
        transform shapes using transformations data from wsireg

        Parameters
        ----------
        transformations
            RegTransformSeq parsable object
        px_idx: bool
            whether shape points are specified in physical coordinates (i.e., microns) or
            in pixel indices
        output_idx: bool
            whether transformed shape points should be output in physical coordinates (i.e., microns) or
            in pixel indices
        """
        if isinstance(transformations, (str, Path, dict)):
            transformations_seq = RegTransformSeq(transformations)
        else:
            transformations_seq = transformations

        invert_nonrigid_transforms(
            transformations_seq.reg_transforms_itk_order
        )

        self.transformed_shape_data = transform_shapes(
            self.shape_data,
            transformations_seq.reg_transforms_itk_order,
            px_idx=px_idx,
            source_res=self.source_res,
            output_idx=output_idx,
            target_res=transformations_seq.output_spacing[0],
        )

    def save_shape_data(self, output_fp, transformed=True):
        """
        Save shape file to GeoJSON on disk.

        Parameters
        ----------
        output_fp: str
            path to the .json file where shape data will be saved
        transformed:bool
            save the transformed shape data or shape data as currently help in memory
        """
        if transformed is True:
            # updated GeoJSON with transformed points
            out_shapes = insert_transformed_pts_gj(
                self.shape_data_gj, self.transformed_shape_data
            )
        else:
            out_shapes = self.shape_data_gj

        json.dump(
            out_shapes,
            open(
                output_fp,
                "w",
            ),
            indent=1,
        )

    def draw_mask(
        self,
        output_size: Tuple[int, int],
        transformed: bool = False,
        labels: bool = False,
    ):
        if labels:
            im_dtype = get_int_dtype(self.n_shapes)
        else:
            im_dtype = np.uint8

        mask = np.zeros(output_size[::-1], dtype=im_dtype)

        if transformed:
            shapes = self.transformed_shape_data
        else:
            shapes = self.shape_data

        # if all([sh["name"] for sh in self.shape_data]):

        for idx, sh in enumerate(shapes):
            mask = cv2.fillPoly(
                mask,
                pts=[sh["array"].astype(np.int32)],
                color=idx + 1 if labels else np.iinfo(im_dtype).max,
            )

        return mask



import os
from copy import deepcopy

import numpy as np
import pytest

from wsireg.reg_shapes import RegShapes

HERE = os.path.dirname(__file__)
GEOJSON_FP = os.path.join(HERE, "fixtures/polygons.geojson")


@pytest.mark.usefixtures("complex_transform")
def test_RegShapes_transform(complex_transform):
    rs = RegShapes(GEOJSON_FP)
    shape0 = deepcopy(rs.shape_data[0])
    rs.transform_shapes(complex_transform)
    shape0_tform = rs.transformed_shape_data[0]
    assert np.array_equal(shape0["array"], shape0_tform["array"]) is False



import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import SimpleITK as sitk

from wsireg.reg_transform import RegTransform
from wsireg.utils.tform_utils import ELX_TO_ITK_INTERPOLATORS


class RegTransformSeq:
    reg_transforms: List[RegTransform] = []
    resampler: Optional[sitk.ResampleImageFilter] = None
    composed_linear_mats: Optional[Dict[str, np.ndarray]] = None
    reg_transforms_itk_order: List[RegTransform] = []

    def __init__(
        self,
        reg_transforms: Optional[
            Union[str, Path, Dict[str, List[str]]]
        ] = None,
        transform_seq_idx: Optional[List[int]] = None,
    ) -> None:

        self._transform_seq_idx = []

        if reg_transforms:
            self.add_transforms(
                reg_transforms, transform_seq_idx=transform_seq_idx
            )
        else:
            self._composite_transform = None
            self._n_transforms = 0

    def add_transforms(self, transforms, transform_seq_idx=None):
        if isinstance(transforms, (str, Path, dict)):
            tform_list, tform_idx = read_wsireg_transform(transforms)
            self.transform_seq_idx = tform_idx
            reg_transforms = [RegTransform(t) for t in tform_list]
            self.reg_transforms = self.reg_transforms + reg_transforms

        elif isinstance(transforms, (list, RegTransform)):
            if isinstance(transforms, RegTransform):
                transforms = [transforms]
            self.reg_transforms = self.reg_transforms + transforms
            self.transform_seq_idx = transform_seq_idx

        self._update_transform_properties()

    @property
    def composite_transform(self) -> sitk.CompositeTransform:
        return self._composite_transform

    @composite_transform.setter
    def composite_transform(self, transforms):
        self._composite_transform = transforms

    @property
    def transform_seq_idx(self) -> List[int]:
        return self._transform_seq_idx

    @transform_seq_idx.setter
    def transform_seq_idx(self, transform_seq):
        if len(self._transform_seq_idx) > 0:
            reindex_val = np.max(self._transform_seq_idx) + 1
        else:
            reindex_val = 0

        transform_seq = [x + reindex_val for x in transform_seq]
        self._transform_seq_idx = self._transform_seq_idx + transform_seq

    @property
    def n_transforms(self) -> int:
        return self._n_transforms

    @n_transforms.setter
    def n_transforms(self) -> None:
        self._n_transforms = len(self.reg_transforms)

    @property
    def output_size(self) -> Tuple[int, int]:
        return self._output_size

    @output_size.setter
    def output_size(self, new_size: Tuple[int, int]) -> None:
        self._output_size = new_size

    @property
    def output_spacing(self) -> Union[Tuple[float, float], Tuple[int, int]]:
        return self._output_spacing

    @output_size.setter
    def output_size(
        self, new_spacing: Union[Tuple[float, float], Tuple[int, int]]
    ) -> None:
        self._output_spacing = new_spacing

    def set_output_spacing(
        self, spacing: Union[Tuple[float, float], Tuple[int, int]]
    ) -> None:

        output_size_scaling = np.asarray(self._output_spacing) / np.asarray(
            spacing
        )
        new_output_size = np.ceil(
            np.multiply(self._output_size, output_size_scaling)
        )
        new_output_size = tuple([int(i) for i in new_output_size])

        self._output_spacing = spacing
        self._output_size = new_output_size

        self._build_resampler()

    def _update_transform_properties(self) -> None:
        self._output_size = self.reg_transforms[-1].output_size
        self._output_spacing = self.reg_transforms[-1].output_spacing
        self._build_transform_data()

    def _build_transform_data(self) -> None:
        self._build_composite_transform(
            self.reg_transforms, self.transform_seq_idx
        )
        self._build_resampler()

    def _build_composite_transform(
        self, reg_transforms, reg_transform_seq_idx
    ) -> None:

        composite_index = []
        for unique_idx in np.unique(reg_transform_seq_idx):
            in_seq_tform_idx = np.where(reg_transform_seq_idx == unique_idx)[0]
            if len(in_seq_tform_idx) > 1:
                composite_index = composite_index + list(
                    in_seq_tform_idx[::-1]
                )
            else:
                composite_index = composite_index + list(in_seq_tform_idx)

        composite_transform = sitk.CompositeTransform(2)

        for tform_idx in composite_index:
            composite_transform.AddTransform(
                reg_transforms[tform_idx].itk_transform
            )

        self._composite_transform = composite_transform
        self.reg_transforms_itk_order = [
            self.reg_transforms[i] for i in composite_index
        ]

    def _build_resampler(self) -> None:
        resampler = sitk.ResampleImageFilter()
        resampler.SetOutputOrigin(self.reg_transforms[-1].output_origin)
        resampler.SetOutputDirection(self.reg_transforms[-1].output_direction)
        resampler.SetSize(self.output_size)
        resampler.SetOutputSpacing(self.output_spacing)

        interpolator = ELX_TO_ITK_INTERPOLATORS.get(
            self.reg_transforms[-1].resample_interpolator
        )
        resampler.SetInterpolator(interpolator)
        resampler.SetTransform(self.composite_transform)
        self.resampler = resampler

    def transform_points(
        self, pt_data: np.ndarray, px_idx=True, source_res=1, output_idx=True
    ) -> np.ndarray:
        tformed_pts = []
        for pt in pt_data:
            if px_idx is True:
                pt = pt * source_res
            for idx, t in enumerate(self.reg_transforms):
                if idx == 0:
                    t_pt = t.inverse_transform.TransformPoint(pt)
                else:
                    t_pt = t.inverse_transform.TransformPoint(t_pt)
            t_pt = np.array(t_pt)

            if output_idx is True:
                t_pt *= 1 / self._output_spacing[0]
            tformed_pts.append(t_pt)

        return np.stack(tformed_pts)

    def append(self, other) -> None:
        self.add_transforms(other.reg_transforms, other.transform_seq_idx)

    def write_transforms(self, output_path: Union[str, Path]):
        return


def read_wsireg_transform(
    parameter_data: Union[str, Path, Dict[Any, Any]]
) -> Tuple[List[Dict[str, List[str]]], List[int]]:

    if isinstance(parameter_data, (str, Path)):
        parameter_data_in = json.load(open(parameter_data, "r"))
    else:
        parameter_data_in = parameter_data

    transform_list = []
    transform_list_seq_id = []

    seq_idx = 0
    for k, v in parameter_data_in.items():
        if k == "initial":
            if isinstance(v, dict):
                transform_list.append(v)
                transform_list_seq_id.append(seq_idx)
                seq_idx += 1
            elif isinstance(v, list):
                for init_tform in v:
                    transform_list.append(init_tform)
                    transform_list_seq_id.append(seq_idx)
                    seq_idx += 1
        else:
            if isinstance(v, dict):
                transform_list.append(v)
                transform_list_seq_id.append(seq_idx)
                seq_idx += 1
            elif isinstance(v, list):
                for tform in v:
                    transform_list.append(tform)
                    transform_list_seq_id.append(seq_idx)
                seq_idx += 1

    return transform_list, transform_list_seq_id



import os

import numpy as np
import pytest

from wsireg.parameter_maps.preprocessing import ImagePreproParams
from wsireg.reg_images.loader import reg_image_loader

# private data logic borrowed from https://github.com/cgohlke/tifffile/tests/test_tifffile.py
HERE = os.path.dirname(__file__)
PRIVATE_DIR = os.path.join(HERE, "private_data")

SKIP_PRIVATE = False
REASON = "private data"

if not os.path.exists(PRIVATE_DIR):
    SKIP_PRIVATE = True


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_czi_read_rgb():
    image_fp = os.path.join(PRIVATE_DIR, "czi_rgb.czi")
    ri = reg_image_loader(image_fp, 1)
    assert len(ri.im_dims) == 3
    assert ri.im_dims[2] == 3
    assert ri.im_dtype == np.uint8
    assert ri.is_rgb is True


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_czi_read_rgb_default_preprocess():
    image_fp = os.path.join(PRIVATE_DIR, "czi_rgb.czi")
    ri = reg_image_loader(image_fp, 1)
    ri.read_reg_image()
    assert ri.reg_image.GetNumberOfComponentsPerPixel() == 1


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_czi_read_rgb_bf_preprocess():
    image_fp = os.path.join(PRIVATE_DIR, "czi_rgb.czi")
    preprocessing = {"image_type": "BF"}
    ri = reg_image_loader(image_fp, 1, preprocessing=preprocessing)
    ri.read_reg_image()
    assert ri.reg_image.GetNumberOfComponentsPerPixel() == 1


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_czi_read_rgb_bf_preprocess_default():
    image_fp = os.path.join(PRIVATE_DIR, "czi_rgb.czi")
    preprocessing = ImagePreproParams()
    ri = reg_image_loader(image_fp, 1, preprocessing=preprocessing)
    ri.read_reg_image()
    assert ri.reg_image.GetNumberOfComponentsPerPixel() == 1


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_czi_read_mc():
    image_fp = os.path.join(PRIVATE_DIR, "czi_4ch_16bit.czi")
    ri = reg_image_loader(image_fp, 1)
    assert len(ri.im_dims) == 3
    assert ri.im_dims[0] == 4
    assert ri.im_dims[2] > 3
    assert ri.im_dtype == np.uint16
    assert ri.is_rgb is False


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_czi_read_mc_default_preprocess():
    image_fp = os.path.join(PRIVATE_DIR, "czi_4ch_16bit.czi")
    ri = reg_image_loader(image_fp, 1)
    ri.read_reg_image()
    assert ri.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert ri.reg_image.GetPixelID() == 1


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_czi_read_mc_fl_preprocess():
    image_fp = os.path.join(PRIVATE_DIR, "czi_4ch_16bit.czi")
    preprocessing = {"image_type": "FL", "as_uint8": True}
    ri = reg_image_loader(image_fp, 1, preprocessing=preprocessing)
    ri.read_reg_image()
    assert ri.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert ri.reg_image.GetPixelID() == 1


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_czi_read_mc_std_preprocess():
    image_fp = os.path.join(PRIVATE_DIR, "czi_4ch_16bit.czi")
    preprocessing = ImagePreproParams()
    ri = reg_image_loader(image_fp, 1, preprocessing=preprocessing)
    ri.read_reg_image()
    assert ri.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert ri.reg_image.GetPixelID() == 1


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_czi_read_mc_selectch_preprocess_list():
    image_fp = os.path.join(PRIVATE_DIR, "czi_4ch_16bit.czi")
    preprocessing = {"ch_indices": [0]}
    ri = reg_image_loader(image_fp, 1, preprocessing=preprocessing)
    ri.read_reg_image()
    assert ri.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert ri.reg_image.GetPixelID() == 1


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_czi_read_mc_selectch_preprocess_int():
    image_fp = os.path.join(PRIVATE_DIR, "czi_4ch_16bit.czi")
    preprocessing = {"ch_indices": 0}
    ri = reg_image_loader(image_fp, 1, preprocessing=preprocessing)
    ri.read_reg_image()
    assert ri.reg_image.GetNumberOfComponentsPerPixel() == 1
    assert ri.reg_image.GetPixelID() == 1


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_czi_read_mc_read_channels():
    image_fp = os.path.join(PRIVATE_DIR, "czi_4ch_16bit.czi")
    ri = reg_image_loader(image_fp, 1)
    ch0 = ri.read_single_channel(0)
    ch1 = ri.read_single_channel(1)
    ch2 = ri.read_single_channel(2)
    ch3 = ri.read_single_channel(3)

    assert np.squeeze(ch0).shape == ri.im_dims[1:]
    assert np.squeeze(ch1).shape == ri.im_dims[1:]
    assert np.squeeze(ch2).shape == ri.im_dims[1:]
    assert np.squeeze(ch3).shape == ri.im_dims[1:]
    assert np.ndim(ch0) == 6
    assert np.ndim(ch1) == 6
    assert np.ndim(ch2) == 6
    assert np.ndim(ch3) == 6
    assert np.mean(ch1) > 0
    assert np.mean(ch2) > 0
    assert np.mean(ch3) > 0
    assert np.array_equal(ch0, ch1) is False
    assert np.array_equal(ch0, ch2) is False
    assert np.array_equal(ch0, ch3) is False
    assert np.array_equal(ch1, ch2) is False
    assert np.array_equal(ch1, ch3) is False
    assert np.array_equal(ch2, ch3) is False
    assert ch0.dtype == np.uint16
    assert ch1.dtype == np.uint16
    assert ch2.dtype == np.uint16
    assert ch3.dtype == np.uint16


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_czi_read_rgb_read_channels():
    image_fp = os.path.join(PRIVATE_DIR, "czi_rgb.czi")
    ri = reg_image_loader(image_fp, 1)
    ch0 = ri.read_single_channel(0)
    ch1 = ri.read_single_channel(1)
    ch2 = ri.read_single_channel(2)

    assert np.squeeze(ch0).shape == ri.im_dims[:2]
    assert np.squeeze(ch1).shape == ri.im_dims[:2]
    assert np.squeeze(ch2).shape == ri.im_dims[:2]
    assert np.ndim(ch0) == 6
    assert np.ndim(ch1) == 6
    assert np.ndim(ch2) == 6
    assert np.array_equal(ch0, ch1) is False
    assert np.array_equal(ch0, ch2) is False
    assert np.array_equal(ch1, ch2) is False
    assert ch0.dtype == np.uint8
    assert ch1.dtype == np.uint8
    assert ch2.dtype == np.uint8


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_scn_read_rgb():
    image_fp = os.path.join(PRIVATE_DIR, "scn_rgb.scn")
    ri = reg_image_loader(image_fp, 1)
    assert len(ri.im_dims) == 3
    assert ri.im_dims[2] == 3
    assert ri.im_dtype == np.uint8
    assert ri.is_rgb is True


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_scn_read_rgb_default_preprocess():
    image_fp = os.path.join(PRIVATE_DIR, "scn_rgb.scn")
    ri = reg_image_loader(image_fp, 1)
    ri.read_reg_image()
    assert ri.reg_image.GetNumberOfComponentsPerPixel() == 1


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_scn_read_rgb_bf_preprocess():
    image_fp = os.path.join(PRIVATE_DIR, "scn_rgb.scn")
    preprocessing = {"image_type": "BF"}
    ri = reg_image_loader(image_fp, 1, preprocessing=preprocessing)
    ri.read_reg_image()
    assert ri.reg_image.GetNumberOfComponentsPerPixel() == 1


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_huron_read_rgb():
    image_fp = os.path.join(PRIVATE_DIR, "huron_rgb.tif")
    ri = reg_image_loader(image_fp, 1)
    assert len(ri.im_dims) == 3
    assert ri.im_dims[2] == 3
    assert ri.im_dtype == np.uint8
    assert ri.is_rgb is True


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_huron_read_rgb_default_preprocess():
    image_fp = os.path.join(PRIVATE_DIR, "huron_rgb.tif")
    ri = reg_image_loader(image_fp, 1)
    ri.read_reg_image()
    assert ri.reg_image.GetNumberOfComponentsPerPixel() == 1


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_huron_read_rgb_bf_preprocess():
    image_fp = os.path.join(PRIVATE_DIR, "huron_rgb.tif")
    preprocessing = {"image_type": "BF"}
    ri = reg_image_loader(image_fp, 1, preprocessing=preprocessing)
    ri.read_reg_image()
    assert ri.reg_image.GetNumberOfComponentsPerPixel() == 1


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_huron_read_rgb_bf_preprocess():
    image_fp = os.path.join(PRIVATE_DIR, "huron_rgb.tif")
    preprocessing = {"image_type": "BF"}
    ri = reg_image_loader(image_fp, 1, preprocessing=preprocessing)
    ri.read_reg_image()
    assert ri.reg_image.GetNumberOfComponentsPerPixel() == 1


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_ometiff_read_rgb():
    image_fp = os.path.join(PRIVATE_DIR, "czi_rgb.ome.tiff")
    ri = reg_image_loader(image_fp, 1)
    assert len(ri.im_dims) == 3
    assert ri.im_dtype == np.uint8
    assert ri.is_rgb is True


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_ometiff_read_rgb_default_preprocess():
    image_fp = os.path.join(PRIVATE_DIR, "czi_rgb.ome.tiff")
    ri = reg_image_loader(image_fp, 1)
    ri.read_reg_image()
    assert ri.reg_image.GetNumberOfComponentsPerPixel() == 1


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_ometiff_read_mc():
    image_fp = os.path.join(PRIVATE_DIR, "czi_4ch_16bit.ome.tiff")
    ri = reg_image_loader(image_fp, 1)

    assert len(ri.im_dims) == 3
    assert ri.im_dims[0] == 4
    assert ri.im_dims[2] > 3
    assert ri.im_dtype == np.uint16
    assert ri.is_rgb is False


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_ometiff_read_mc_default_preprocess():
    image_fp = os.path.join(PRIVATE_DIR, "czi_4ch_16bit.ome.tiff")
    ri = reg_image_loader(image_fp, 1)
    ri.read_reg_image()
    assert ri.reg_image.GetNumberOfComponentsPerPixel() == 1


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_czi_read_mc_read_channels():
    image_fp = os.path.join(PRIVATE_DIR, "czi_4ch_16bit.ome.tiff")
    ri = reg_image_loader(image_fp, 1)
    ch0 = ri.read_single_channel(0)
    ch1 = ri.read_single_channel(1)
    ch2 = ri.read_single_channel(2)
    ch3 = ri.read_single_channel(3)

    assert np.squeeze(ch0).shape == ri.im_dims[1:]
    assert np.squeeze(ch1).shape == ri.im_dims[1:]
    assert np.squeeze(ch2).shape == ri.im_dims[1:]
    assert np.squeeze(ch3).shape == ri.im_dims[1:]
    assert np.ndim(ch0) == 2
    assert np.ndim(ch1) == 2
    assert np.ndim(ch2) == 2
    assert np.ndim(ch3) == 2
    assert np.array_equal(ch0, ch1) is False
    assert np.array_equal(ch0, ch2) is False
    assert np.array_equal(ch0, ch3) is False
    assert np.array_equal(ch1, ch2) is False
    assert np.array_equal(ch1, ch3) is False
    assert np.array_equal(ch2, ch3) is False
    assert ch0.dtype == np.uint16
    assert ch1.dtype == np.uint16
    assert ch2.dtype == np.uint16
    assert ch3.dtype == np.uint16


@pytest.mark.skipif(SKIP_PRIVATE, reason=REASON)
def test_czi_read_rgb_read_channels():
    image_fp = os.path.join(PRIVATE_DIR, "czi_rgb.ome.tiff")
    ri = reg_image_loader(image_fp, 1)
    ch0 = ri.read_single_channel(0)
    ch1 = ri.read_single_channel(1)
    ch2 = ri.read_single_channel(2)

    assert np.squeeze(ch0).shape == ri.im_dims[:2]
    assert np.squeeze(ch1).shape == ri.im_dims[:2]
    assert np.squeeze(ch2).shape == ri.im_dims[:2]
    assert np.ndim(ch0) == 2
    assert np.ndim(ch1) == 2
    assert np.ndim(ch2) == 2
    assert np.array_equal(ch0, ch1) is False
    assert np.array_equal(ch0, ch2) is False
    assert np.array_equal(ch1, ch2) is False
    assert ch0.dtype == np.uint8
    assert ch1.dtype == np.uint8
    assert ch2.dtype == np.uint8



import json
from pathlib import Path
from typing import Dict, List, Union

import itk
import numpy as np
import SimpleITK as sitk

from wsireg.parameter_maps.reg_model import RegModel
from wsireg.utils.itk_im_conversions import itk_image_to_sitk_image

NP_TO_SITK_DTYPE = {
    np.dtype(np.int8): 0,
    np.dtype(np.uint8): 1,
    np.dtype(np.int16): 2,
    np.dtype(np.uint16): 3,
    np.dtype(np.int32): 4,
    np.dtype(np.uint32): 5,
    np.dtype(np.int64): 6,
    np.dtype(np.uint64): 7,
    np.dtype(np.float32): 8,
    np.dtype(np.float64): 9,
    np.dtype(np.complex64): 10,
    np.dtype(np.complex64): 11,
}


def sitk_pmap_to_dict(pmap):
    """
    Convert SimpleElastix ParameterMap to python dictionary

    Parameters
    ----------
    pmap
        SimpleElastix ParameterMap

    Returns
    -------
    Python dict of SimpleElastix ParameterMap
    """
    pmap_dict = {}
    for k, v in pmap.items():
        if k in ["image", "invert"]:
            t_pmap = {}
            for k2, v2 in v.items():
                t_pmap[k2] = v2
            pmap_dict[k] = t_pmap
        else:
            pmap_dict[k] = v
    return pmap_dict


def pmap_dict_to_sitk(pmap_dict):
    """
    Convert python dict to SimpleElastix ParameterMap

    Parameters
    ----------
    pmap_dict
        SimpleElastix ParameterMap in python dictionary

    Returns
    -------
    SimpleElastix ParameterMap of Python dict
    """
    # pmap = sitk.ParameterMap()
    # pmap = {}
    # for k, v in pmap_dict.items():
    #     pmap[k] = v
    return pmap_dict


def pmap_dict_to_json(pmap_dict, output_file):
    """
    Save python dict of ITKElastix to json

    Parameters
    ----------
    pmap_dict : dict
        parameter map stored in python dict
    output_file : str
        filepath of where to save the json
    """
    with open(output_file, "w") as fp:
        json.dump(pmap_dict, fp, indent=4)


def json_to_pmap_dict(json_file):
    """
    Load python dict of SimpleElastix stored in json

    Parameters
    ----------
    json_file : dict
        filepath to json contained SimpleElastix parameter map
    """
    with open(json_file, "r") as fp:
        pmap_dict = json.load(fp)
    return pmap_dict


def _prepare_reg_models(
    reg_params: List[Union[RegModel, Dict[str, List[str]]]]
) -> List[Dict[str, List[str]]]:
    prepared_params = []
    for rp in reg_params:
        if isinstance(rp, RegModel):
            prepared_params.append(rp.value)
        elif isinstance(rp, str):
            prepared_params.append(RegModel[rp].value)
        elif isinstance(rp, dict):
            prepared_params.append(rp)
    return prepared_params


def parameter_to_itk_pobj(reg_param_map):
    """
    Transfer parameter data stored in dict to ITKElastix ParameterObject

    Parameters
    ----------
    reg_param_map: dict
        elastix registration parameters

    Returns
    -------
    itk_param_map:itk.ParameterObject
        ITKElastix object for registration parameters
    """
    parameter_object = itk.ParameterObject.New()
    itk_param_map = parameter_object.GetDefaultParameterMap("rigid")
    for k, v in reg_param_map.items():
        itk_param_map[k] = v
    return itk_param_map


def register_2d_images_itkelx(
    source_image,
    target_image,
    reg_params: List[Dict[str, List[str]]],
    reg_output_fp: Union[str, Path],
    histogram_match=False,
    return_image=False,
):
    """
    Register 2D images with multiple models and return a list of elastix
    transformation maps.

    Parameters
    ----------
    source_image : SimpleITK.Image
        RegImage of image to be aligned
    target_image : SimpleITK.Image
        RegImage that is being aligned to (grammar is hard)
    reg_params : list of dict
        registration parameter maps stored in a dict, can be file paths to SimpleElastix parameterMaps stored
        as text or one of the default parameter maps (see parameter_load() function)
    reg_output_fp : str
        where to store registration outputs (iteration data and transformation files)
    histogram_match : bool
        whether to attempt histogram matching to improve registration
    Returns
    -------
        tform_list: list
            list of ITKElastix transformation parameter maps
        image: itk.Image
            resulting registered moving image
    """
    if histogram_match is True:
        matcher = sitk.HistogramMatchingImageFilter()
        matcher.SetNumberOfHistogramLevels(64)
        matcher.SetNumberOfMatchPoints(7)
        matcher.ThresholdAtMeanIntensityOn()
        source_image.image = matcher.Execute(
            source_image.reg_image, target_image.reg_image
        )

    pixel_id = source_image.reg_image.GetPixelID()
    source_image.reg_image_sitk_to_itk()
    target_image.reg_image_sitk_to_itk()

    selx = itk.ElastixRegistrationMethod.New(
        source_image.reg_image, target_image.reg_image
    )

    # Set additional options
    selx.SetLogToConsole(True)
    selx.SetOutputDirectory(str(reg_output_fp))

    if source_image.mask is not None:
        selx.SetMovingMask(source_image.mask)

    if target_image.mask is not None:
        selx.SetFixedMask(target_image.mask)

    selx.SetMovingImage(source_image.reg_image)
    selx.SetFixedImage(target_image.reg_image)

    parameter_object_registration = itk.ParameterObject.New()
    for idx, pmap in enumerate(reg_params):
        if idx == 0:
            pmap["WriteResultImage"] = ["true"] if return_image else ["false"]
            if target_image.mask is not None:
                pmap["AutomaticTransformInitialization"] = ["false"]
            else:
                pmap["AutomaticTransformInitialization"] = ['true']

            parameter_object_registration.AddParameterMap(pmap)
        else:
            pmap["WriteResultImage"] = ["true"] if return_image else ["false"]
            pmap["AutomaticTransformInitialization"] = ['false']
            parameter_object_registration.AddParameterMap(pmap)

    selx.SetParameterObject(parameter_object_registration)

    # Update filter object (required)
    selx.UpdateLargestPossibleRegion()

    # Results of Registration
    result_transform_parameters = selx.GetTransformParameterObject()

    # execute registration:
    tform_list = []
    for idx in range(result_transform_parameters.GetNumberOfParameterMaps()):
        tform = {}
        for k, v in result_transform_parameters.GetParameterMap(idx).items():
            tform[k] = v
        tform_list.append(tform)

    if return_image is False:
        return tform_list
    else:
        image = selx.GetOutput()
        image = itk_image_to_sitk_image(image)
        image = sitk.Cast(image, pixel_id)
        return tform_list, image


#
# def register_2d_images(
#     source_image,
#     target_image,
#     reg_params,
#     reg_output_fp,
#     histogram_match=False,
#     compute_inverse=True,
#     return_image=False,
# ):
#     """
#     Register 2D images with multiple models and return a list of elastix
#     transformation maps.
#
#     Parameters
#     ----------
#     source_image : SimpleITK.Image
#         RegImage of image to be aligned
#     target_image : SimpleITK.Image
#         RegImage that is being aligned to (grammar is hard)
#     reg_params : dict
#         registration parameter maps stored in a dict, can be file paths to SimpleElastix parameterMaps stored
#         as text or one of the default parameter maps (see parameter_load() function)
#     reg_output_fp : str
#         where to store registration outputs (iteration data and transformation files)
#     histogram_match : bool
#         whether to attempt histogram matching to improve registration
#     compute_inverse : bool
#         whether to compute the inverse for BSplineTransforms. This is needed to transform
#         point sets
#     Returns
#     -------
#     list of SimpleElastix transformation parameter maps
#
#     """
#
#     try:
#         selx = sitk.SimpleElastix()
#     except AttributeError:
#         selx = sitk.ElastixImageFilter()
#
#     selx.SetOutputDirectory(str(reg_output_fp))
#
#     # these parameters may be made optional later
#     # not critical though
#     if histogram_match is True:
#         matcher = sitk.HistogramMatchingImageFilter()
#         matcher.SetNumberOfHistogramLevels(64)
#         matcher.SetNumberOfMatchPoints(7)
#         matcher.ThresholdAtMeanIntensityOn()
#         source_image.image = matcher.Execute(
#             source_image.image, target_image.image
#         )
#
#     selx.SetMovingImage(source_image.image)
#     selx.SetFixedImage(target_image.image)
#
#     if source_image.mask is not None:
#         selx.SetMovingMask(source_image.mask)
#
#     if target_image.mask is not None:
#         selx.SetFixedMask(target_image.mask)
#
#     for idx, reg_param in enumerate(reg_params):
#         if idx == 0:
#             pmap = parameter_load(reg_param)
#             pmap["WriteResultImage"] = ("false",)
#             if target_image.mask is not None:
#                 pmap["AutomaticTransformInitialization"] = ("false",)
#             selx.SetParameterMap(pmap)
#         else:
#             pmap = parameter_load(reg_param)
#             pmap["WriteResultImage"] = ("false",)
#             selx.AddParameterMap(pmap)
#
#     selx.LogToConsoleOn()
#     selx.LogToFileOn()
#
#     # execute registration:
#     if return_image is False:
#         selx.Execute()
#     else:
#         image = selx.Execute()
#
#     tform_list = list(selx.GetTransformParameterMap())
#
#     if compute_inverse is True:
#         compute_tforms = []
#         for idx, reg_param in enumerate(reg_params):
#             pmap = parameter_load(reg_param)
#             pmap["WriteResultImage"] = ("false",)
#             if pmap["Transform"][0] == "BSplineTransform":
#                 compute_tforms.append((idx, pmap))
#
#         if len(compute_tforms) > 0:
#             for idx, compute_tform in compute_tforms:
#
#                 selx.SetMovingImage(target_image.image)
#                 selx.SetFixedImage(target_image.image)
#                 compute_tform["Metric"] = ["DisplacementMagnitudePenalty"]
#                 max_step_double = compute_tform["MaximumStepLength"]
#                 compute_tform["MaximumStepLength"] = [
#                     str(float(step) * 2) for step in max_step_double
#                 ]
#                 selx.SetParameterMap(compute_tform)
#
#                 with tempfile.TemporaryDirectory() as tempdir:
#                     temp_tform_path = Path(tempdir) / "temp_tform.txt"
#                     tform_out = list(selx.GetTransformParameterMap())[-1]
#                     tform_out["InitialTransformParametersFileName"] = [
#                         "NoInitialTransform"
#                     ]
#                     sitk.WriteParameterFile(
#                         tform_out,
#                         str(temp_tform_path),
#                     )
#                     selx.SetInitialTransformParameterFileName(
#                         str(temp_tform_path)
#                     )
#                     selx.SetOutputDirectory(tempdir)
#                     selx.Execute()
#                     inverted_tform = list(selx.GetTransformParameterMap())[0]
#
#                 tform_normal = tform_list[idx]
#                 tform_list[idx] = {
#                     "image": tform_normal,
#                     "invert": inverted_tform,
#                 }
#         else:
#             raise ValueError(
#                 "No support for inverting intermediate BSplineTransforms"
#             )
#     else:
#         print("no inversions to compute")
#
#     if return_image is False:
#         return tform_list
#     else:
#         image = sitk.Cast(image, source_image.image.GetPixelID())
#         return tform_list, image



from enum import Enum
from typing import Callable, Dict, List, NamedTuple, Optional, Tuple, Union

from pydantic import BaseModel, validator


class ImageType(str, Enum):
    """Set the photometric interpretation of the image
    * "FL": background is black (fluorescence)
    * "BF": Background is white (brightfield)
    """

    DARK = "FL"
    LIGHT = "BF"


class CoordinateFlip(str, Enum):
    """Coordinate flip options
    * "h" : horizontal flip
    * "v" : vertical flip
    """

    HORIZONTAL = "h"
    VERTIAL = "v"


class BoundingBox(NamedTuple):
    X: int
    Y: int
    WIDTH: int
    HEIGHT: int


def _transform_to_bbox(mask_bbox: Union[Tuple[int, int, int, int], List[int]]):
    return BoundingBox(*mask_bbox)


def _index_to_list(ch_indices: Union[int, List[int]]):
    if isinstance(ch_indices, int):
        ch_indices = [ch_indices]
    return ch_indices


def _transform_custom_proc(
    custom_procs: Union[List[Callable], Tuple[Callable, ...]]
):
    return {
        f"custom processing {str(idx+1).zfill(2)}": proc
        for idx, proc in enumerate(custom_procs)
    }


class ImagePreproParams(BaseModel):
    """Preprocessing parameter model

    Attributes
    ----------
    image_type: ImageType
        Whether image is dark or light background. Light background images are intensity inverted
        by default
    max_int_proj: bool
        Perform max intensity projection number of channels > 1.
    contrast_enhance: bool
        Enhance contrast of image
    invert_intensity: bool
        invert the intensity of an image
    rot_cc: int, float
        Rotate image counter-clockwise by degrees, can be positive or negative (cw rot)
    flip: CoordinateFlip, default: None
        flip coordinates, "v" = vertical flip, "h" = horizontal flip
    crop_to_mask_bbox: bool
        Convert a binary mask to a bounding box and crop to this area
    mask_bbox: tuple or list of 4 ints
        supply a pre-computed list of bbox info of form x,y,width,height
    """

    # intensity preprocessing
    image_type: ImageType = ImageType.DARK
    max_int_proj: bool = True
    ch_indices: Optional[List[int]] = None
    as_uint8: bool = True
    contrast_enhance: bool = False
    invert_intensity: bool = False
    custom_processing: Optional[Dict[str, Callable]] = None

    # spatial preprocessing
    rot_cc: Union[int, float] = 0
    flip: Optional[CoordinateFlip] = None
    crop_to_mask_bbox: bool = False
    mask_bbox: Optional[BoundingBox] = None
    downsampling: int = 1
    use_mask: bool = True

    @validator('mask_bbox', pre=True)
    def _make_bbox(cls, v):
        return _transform_to_bbox(v)

    @validator('ch_indices', pre=True)
    def _make_ch_list(cls, v):
        return _index_to_list(v)

    @validator('custom_processing', pre=True)
    def _check_custom_prepro(cls, v):
        if isinstance(v, (list, tuple)):
            return _transform_custom_proc(v)
        else:
            return v
